"""
test_moment_tensor_conversion.py
*****************

Tests for src/convert/moment_tensor_conversion.py
"""

import unittest
import sys

import numpy as np

from MTfit.convert import moment_tensor_conversion as mtc
from MTfit.convert import MT33_MT6
from MTfit.convert import MT6_MT33
from MTfit.convert import MT6_TNPE
from MTfit.convert import MT6_Tape
from MTfit.convert import MT33_TNPE
from MTfit.convert import MT33_SDR
from MTfit.convert import MT33_GD
from MTfit.convert import E_tk
from MTfit.convert import tk_uv
from MTfit.convert import E_uv
from MTfit.convert import E_GD
from MTfit.convert import GD_basic_cdc
from MTfit.convert import TNP_SDR
from MTfit.convert import TP_FP
from MTfit.convert import FP_SDR
from MTfit.convert import basic_cdc_GD
from MTfit.convert import GD_E
from MTfit.convert import SDR_TNP
from MTfit.convert import SDR_SDR
from MTfit.convert import FP_TNP
from MTfit.convert import SDSD_FP
from MTfit.convert import SDR_FP
from MTfit.convert import SDR_SDSD
from MTfit.convert import FP_SDSD
from MTfit.convert import Tape_MT33
from MTfit.convert import Tape_MT6
from MTfit.convert import Tape_TNPE
from MTfit.convert import normal_SD
from MTfit.convert import toa_vec
from MTfit.convert import output_convert
from MTfit.convert import isotropic_c
from MTfit.convert import MT6_biaxes
from MTfit.convert import MT6c_D6
from MTfit.convert import is_isotropic_c
from MTfit.convert import c21_cvoigt
from MTfit.convert import c_norm
from MTfit.convert import moment_tensor_conversion
from MTfit.utilities import C_EXTENSION_FALLBACK_LOG_MSG
from MTfit.utilities.unittest_utils import TestCase
from MTfit.utilities.unittest_utils import get_extension_skip_if_args

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock


C_EXTENSIONS = get_extension_skip_if_args('MTfit.convert.cmoment_tensor_conversion')


class PythonOnly(object):

    def __enter__(self, *args, **kwargs):
        self.cmoment_tensor_conversion = moment_tensor_conversion.cmoment_tensor_conversion
        moment_tensor_conversion.cmoment_tensor_conversion = False

    def __exit__(self, *args, **kwargs):
        moment_tensor_conversion.cmoment_tensor_conversion = self.cmoment_tensor_conversion


class MomentTensorConvertTestCase(TestCase):

    def setUp(self):
        self.DC6 = np.array([[0.], [0.], [0.], [1.], [0.], [0.]])
        self.MT6 = np.array([[1.], [0.], [-2.], [1.], [0.], [2.]])
        self.MT6norm = np.sqrt(np.sum(self.MT6*self.MT6))
        self.DC33 = np.array([[0, 1/np.sqrt(2), 0], [1/np.sqrt(2), 0, 0], [0, 0, 0]])
        self.MT33 = np.array([[1, 1/np.sqrt(2), 0], [1/np.sqrt(2), 0, 2/np.sqrt(2)], [0, 2/np.sqrt(2), -2]])
        self.MT33norm = np.sqrt(np.sum(self.MT33*self.MT33))
        # Values from matlab code
        self.Tmt = np.matrix([0.78187414360087504, 0.57856722491374268, 0.23223434247330654]).T
        self.Nmt = np.matrix([0.61708838355029516, -0.66519145682767544, -0.42038345905941465]).T
        self.Pmt = np.matrix([0.088739790712409491, -0.47199607, 0.87712311]).T
        self.Emt = np.matrix([1.5232412549475853, 0.23777306063666762, -2.7610143155842528]).T
        self.amt = 1.9822156229199277
        self.vmt = -0.19209745970264186
        self.adc = np.pi/2
        self.vdc = 0.5
        self.Tdc = np.matrix([1/np.sqrt(2), 1/np.sqrt(2), 0]).T
        self.Ndc = np.matrix([0., 0., 1.]).T
        self.Pdc = np.matrix([-1/np.sqrt(2), 1/np.sqrt(2), 0]).T
        self.Edc = np.matrix([1/np.sqrt(2), 0, -1/np.sqrt(2)]).T
        self.Gmt = 0.22691207246521938
        self.Dmt = -0.18360401027891848
        self.Umt = 0.41369317844275666
        self.Vmt = -0.12072857842564184
        self.tmt = 0.41369317844275666
        self.kmt = -0.12072857842564184
        self.Gdc = 0.
        self.Ddc = 0.
        self.Udc = 0.
        self.Vdc = 0.
        self.tdc = 0.
        self.kdc = 0.
        self.N1mt = np.matrix([0.61561701812768677, 0.075357186487224101, 0.78443417916119973]).T
        self.N2mt = np.matrix([0.49012000239468767, 0.7428604317087556, -0.4560052217399625]).T
        self.N1dc = np.matrix([0, 1., 0]).T
        self.N2dc = np.matrix([1., 0, 0]).T
        self.S1mt = 1.69259957
        self.D1mt = 0.66901303
        self.R1mt = -2.31557077
        self.S2mt = 5.69996974
        self.D2mt = 1.09729495
        self.R2mt = -1.07883773
        self.S1dc = 0.
        self.D1dc = 90.*np.pi/180
        self.R1dc = 0.
        self.S2dc = 270.*np.pi/180
        self.D2dc = 90.*np.pi/180
        self.R2dc = -180.*np.pi/180
        self.Kmt = 5.69996974
        self.Hmt = 0.45600522
        self.Omt = -1.07883773
        self.Kdc = 0.
        self.Hdc = 0.
        self.Odc = 0.
        self.T2 = np.array([[1., 0.99931011, 0.99753384, 1., 0.98559856, 1., 1., 1.],
                            [0., 0., 0.0701871, 0., 0.16910198, 0., 0., 0.],
                            [0., 0.0371391, 0., 0., 0., 0., 0., 0.]])
        self.N2 = np.array([[0., 0., -0.0701871, 0., -0.16910198, 0., 0., 0.],
                            [1., 1., 0.99753384, -0.98559856,  0.98559856, 1., 1., 1.],
                            [0., 0., 0., -0.16910198, 0., 0., 0., 0.]])
        self.P2 = np.array([[0., -0.0371391, 0., 0., 0., 0., 0., 0.],
                            [0., 0., 0., 0.16910198, 0., 0., 0., 0.],
                            [1., 0.99931011, 1., -0.98559856, 1., 1., 1., 1.]])
        self.S21 = np.array([90., 90., 94.02473349, 279.5980303, 99.73561032, 90., 90., 90.])*np.pi/180.
        self.S22 = np.array([270., 270., 274.02473349, 80.4019697, 279.73561032, 270., 270., 270.])*np.pi/180.
        self.D21 = np.array([45., 42.87159666, 45., 45.81931182, 45., 45., 45., 45.])*np.pi/180.
        self.D22 = np.array([45., 47.12840334, 45., 45.81931182, 45., 45., 45., 45.])*np.pi/180.
        self.R21 = np.array([-90., -90., -90., -76.36129238206001, -90., -90., -90., -90.])*np.pi/180.
        self.R22 = np.array([-90., -90., -90., -103.63870728, -90., -90., -90., -90.])*np.pi/180.

    def tearDown(self):
        del self.DC6
        del self.MT6
        del self.DC33
        del self.MT33
        del self.MT6norm
        del self.MT33norm
        del self.Tmt
        del self.Nmt
        del self.Pmt
        del self.Emt
        del self.Tdc
        del self.Ndc
        del self.Pdc
        del self.Edc
        del self.amt
        del self.vmt
        del self.adc
        del self.vdc
        del self.Gmt
        del self.Dmt
        del self.Umt
        del self.Vmt
        del self.tmt
        del self.kmt
        del self.Gdc
        del self.Ddc
        del self.Udc
        del self.Vdc
        del self.tdc
        del self.kdc
        del self.N1mt
        del self.N2mt
        del self.N1dc
        del self.N2dc
        del self.S1mt
        del self.D1mt
        del self.R1mt
        del self.S2mt
        del self.D2mt
        del self.R2mt
        del self.S1dc
        del self.D1dc
        del self.R1dc
        del self.S2dc
        del self.D2dc
        del self.R2dc
        del self.Kmt
        del self.Hmt
        del self.Omt
        del self.Kdc
        del self.Hdc
        del self.Odc
        del self.T2
        del self.P2
        del self.N2
        del self.S21
        del self.D21
        del self.R21
        del self.S22
        del self.D22
        del self.R22

    def assertSigmaEquals(self, first, second, *args):
        try:
            return self.assertAlmostEquals(first, second, *args)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(np.abs(first), np.pi, *args)
                return self.assertAlmostEquals(-first, second, *args)
            except Exception:
                raise e1

    def test_MT33_MT6(self):
        DC6 = MT33_MT6(self.DC33)
        MT6 = MT33_MT6(self.MT33)
        self.assertAlmostEqual(DC6, self.DC6)
        self.assertAlmostEqual(MT6, self.MT6/self.MT6norm)

    def test_MT6_MT33(self):
        DC33 = MT6_MT33(self.DC6)
        MT33 = MT6_MT33(self.MT6)
        self.assertAlmostEqual(DC33, self.DC33)
        self.assertAlmostEqual(MT33, self.MT33)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_MT6_TNPE(self, logger):
        with PythonOnly():
            [Tdc, Ndc, Pdc, Edc] = MT6_TNPE(self.DC6)
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            [Tmt, Nmt, Pmt, Emt] = MT6_TNPE(self.MT6)
            self.assertVectorEquals(Tdc, self.Tdc)
            self.assertVectorEquals(Ndc, self.Ndc)
            self.assertVectorEquals(Pdc, self.Pdc)
            self.assertAlmostEquals(Edc, self.Edc)
            self.assertVectorEquals(Tmt, self.Tmt)
            self.assertVectorEquals(Nmt, self.Nmt)
            self.assertVectorEquals(Pmt, self.Pmt)
            self.assertAlmostEquals(Emt, self.Emt)
            [T, N, P, E] = MT6_TNPE(np.append(self.MT6, self.DC6, 1))
            self.assertVectorEquals(T[:, 1], self.Tdc)
            self.assertVectorEquals(N[:, 1], self.Ndc)
            self.assertVectorEquals(P[:, 1], self.Pdc)
            self.assertAlmostEquals(E[:, 1], self.Edc)
            self.assertVectorEquals(T[:, 0], self.Tmt)
            self.assertVectorEquals(N[:, 0], self.Nmt)
            self.assertVectorEquals(P[:, 0], self.Pmt)
            self.assertAlmostEquals(E[:, 0], self.Emt)
            self.assertEqual(logger.info.call_args_list, [mock.call(C_EXTENSION_FALLBACK_LOG_MSG),
                                                          mock.call(C_EXTENSION_FALLBACK_LOG_MSG),
                                                          mock.call(C_EXTENSION_FALLBACK_LOG_MSG)])

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_MT6_TNPE_cython(self, logger):
        [Tdc, Ndc, Pdc, Edc] = MT6_TNPE(self.DC6)
        [Tmt, Nmt, Pmt, Emt] = MT6_TNPE(self.MT6)
        self.assertVectorEquals(Tdc, self.Tdc)
        self.assertVectorEquals(Ndc, self.Ndc)
        self.assertVectorEquals(Pdc, self.Pdc)
        self.assertAlmostEquals(Edc, self.Edc)
        self.assertVectorEquals(Tmt, self.Tmt)
        self.assertVectorEquals(Nmt, self.Nmt)
        self.assertVectorEquals(Pmt, self.Pmt)
        self.assertAlmostEquals(Emt, self.Emt)
        [T, N, P, E] = MT6_TNPE(np.append(self.MT6, self.DC6, 1))
        self.assertVectorEquals(T[:, 1], self.Tdc)
        self.assertVectorEquals(N[:, 1], self.Ndc)
        self.assertVectorEquals(P[:, 1], self.Pdc)
        self.assertAlmostEquals(E[:, 1], self.Edc)
        self.assertVectorEquals(T[:, 0], self.Tmt)
        self.assertVectorEquals(N[:, 0], self.Nmt)
        self.assertVectorEquals(P[:, 0], self.Pmt)
        self.assertAlmostEquals(E[:, 0], self.Emt)
        logger.info.assert_not_called()

    def test_MT6_Tape(self):
        mtc._CYTHON = False
        [Gmt, Dmt, Kmt, Hmt, Omt] = MT6_Tape(self.MT6)
        self.assertAlmostEquals(Gmt, self.Gmt)
        self.assertAlmostEquals(Dmt, self.Dmt)
        self.assertAlmostEquals(Kmt, self.Kmt)
        self.assertAlmostEquals(Hmt, self.Hmt)
        self.assertAlmostEquals(Omt, self.Omt)
        [Gdc, Ddc, Kdc, Hdc, Odc] = MT6_Tape(self.DC6)
        self.assertAlmostEquals(Gdc, self.Gdc)
        self.assertAlmostEquals(Ddc, self.Ddc)
        try:
            self.assertAlmostEquals(Kdc, self.Kdc)
        except Exception:
            self.assertAlmostEquals(Kdc-np.pi, self.Kdc)
        self.assertAlmostEquals(Hdc, self.Hdc)
        self.assertAlmostEquals(Odc, self.Odc)

    def test_MT6_Tape_multiple(self):
        [G, D, K, H, O] = MT6_Tape(np.append(self.MT6, self.DC6, 1))
        self.assertAlmostEquals(G[0], self.Gmt)
        self.assertAlmostEquals(D[0], self.Dmt)
        self.assertAlmostEquals(K[0], self.Kmt)
        self.assertAlmostEquals(H[0], self.Hmt)
        self.assertAlmostEquals(O[0], self.Omt)
        self.assertAlmostEquals(G[1], self.Gdc)
        self.assertAlmostEquals(D[1], self.Ddc)
        try:
            self.assertAlmostEquals(K[1], self.Kdc)
        except Exception:
            self.assertAlmostEquals(K[1]-np.pi, self.Kdc)
        self.assertAlmostEquals(H[1], self.Hdc)
        self.assertAlmostEquals(O[1], self.Odc)

    def test_MT33_TNPE(self):
        [Tdc, Ndc, Pdc, Edc] = MT33_TNPE(self.DC33)
        [Tmt, Nmt, Pmt, Emt] = MT33_TNPE(self.MT33)
        self.assertVectorEquals(Tdc, self.Tdc)
        self.assertVectorEquals(Ndc, self.Ndc)
        self.assertVectorEquals(Pdc, self.Pdc)
        self.assertAlmostEquals(Edc, self.Edc)
        self.assertVectorEquals(Tmt, self.Tmt)
        self.assertVectorEquals(Nmt, self.Nmt)
        self.assertVectorEquals(Pmt, self.Pmt)
        self.assertAlmostEquals(Emt, self.Emt)

    def test_MT33_SDR(self):
        [Sdc, Ddc, Rdc] = MT33_SDR(self.DC33)
        [Smt, Dmt, Rmt] = MT33_SDR(self.MT33)
        try:
            self.assertAlmostEquals(Sdc, self.S1dc)
            self.assertAlmostEquals(Ddc, self.D1dc)
            self.assertAlmostEquals(Rdc, self.R1dc)
        except Exception:
            self.assertAlmostEquals(Sdc, self.S2dc)
            self.assertAlmostEquals(Ddc, self.D2dc)
            self.assertAlmostEquals(Rdc, self.R2dc)
        try:
            self.assertAlmostEquals(Smt, self.S1mt)
            self.assertAlmostEquals(Dmt, self.D1mt)
            self.assertAlmostEquals(Rmt, self.R1mt)
        except Exception:
            self.assertAlmostEquals(Smt, self.S2mt)
            self.assertAlmostEquals(Dmt, self.D2mt)
            self.assertAlmostEquals(Rmt, self.R2mt)

    def test_MT33_GD(self):
        [Gmt, Dmt] = MT33_GD(self.MT33)
        self.assertAlmostEquals(Gmt, self.Gmt)
        self.assertAlmostEquals(Dmt, self.Dmt)

    def test_E_tk(self):
        tdc, kdc = E_tk(self.Edc)
        tmt, kmt = E_tk(self.Emt)
        self.assertAlmostEquals(tdc, self.tdc)
        self.assertAlmostEquals(kdc, self.kdc)
        self.assertAlmostEquals(tmt, self.tmt)
        self.assertAlmostEquals(kmt, self.kmt)

    def test_E_tk_multiple(self):
        t, k = E_tk(np.append(self.Emt, self.Edc, 1))
        self.assertAlmostEquals(t[1], self.tdc)
        self.assertAlmostEquals(k[1], self.kdc)
        self.assertAlmostEquals(t[0], self.tmt)
        self.assertAlmostEquals(k[0], self.kmt)

    def test_tk_uv(self):
        Udc, Vdc = tk_uv(self.tdc, self.kdc)
        Umt, Vmt = tk_uv(self.tmt, self.kmt)
        self.assertAlmostEquals(Udc, self.Udc)
        self.assertAlmostEquals(Vdc, self.Vdc)
        self.assertAlmostEquals(Umt, self.Umt)
        self.assertAlmostEquals(Vmt, self.Vmt)

    def test_tk_uv_multiple(self):
        u, v = tk_uv(np.array([self.tmt, self.tdc]), np.array([self.kmt, self.kdc]))
        self.assertAlmostEquals(u[1], self.Udc)
        self.assertAlmostEquals(v[1], self.Vdc)
        self.assertAlmostEquals(u[0], self.Umt)
        self.assertAlmostEquals(v[0], self.Vmt)

    def test_E_uv(self):
        Udc, Vdc = E_uv(self.Edc)
        Umt, Vmt = E_uv(self.Emt)
        self.assertAlmostEquals(Udc, self.Udc)
        self.assertAlmostEquals(Vdc, self.Vdc)
        self.assertAlmostEquals(Umt, self.Umt)
        self.assertAlmostEquals(Vmt, self.Vmt)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_E_GD_python(self, logger):
        with PythonOnly():
            Gdc, Ddc = E_GD(self.Edc)
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            Gmt, Dmt = E_GD(self.Emt)
            self.assertAlmostEquals(Gdc, self.Gdc)
            self.assertAlmostEquals(Ddc, self.Ddc)
            self.assertAlmostEquals(Gmt, self.Gmt)
            self.assertAlmostEquals(Dmt, self.Dmt)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_E_GD_multiple_python(self, logger):
        with PythonOnly():
            t, k = E_GD(np.append(self.Emt, self.Edc, 1))
            self.assertAlmostEquals(t[1], self.Gdc)
            self.assertAlmostEquals(k[1], self.Ddc)
            self.assertAlmostEquals(t[0], self.Gmt)
            self.assertAlmostEquals(k[0], self.Dmt)
            logger.info.assert_called_with(C_EXTENSION_FALLBACK_LOG_MSG)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_E_GD_cython(self, logger):
        Gdc, Ddc = E_GD(self.Edc)
        Gmt, Dmt = E_GD(self.Emt)
        self.assertAlmostEquals(Gdc, self.Gdc)
        self.assertAlmostEquals(Ddc, self.Ddc)
        self.assertAlmostEquals(Gmt, self.Gmt)
        self.assertAlmostEquals(Dmt, self.Dmt)
        logger.info.assert_not_called()

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_E_GD_multiple_cython(self, logger):
        t, k = E_GD(np.append(self.Emt, self.Edc, 1))
        self.assertAlmostEquals(t[1], self.Gdc)
        self.assertAlmostEquals(k[1], self.Ddc)
        self.assertAlmostEquals(t[0], self.Gmt)
        self.assertAlmostEquals(k[0], self.Dmt)
        logger.info.assert_not_called()

    def test_GD_basic_cdc(self):
        adc, vdc = GD_basic_cdc(self.Gdc, self.Ddc)
        amt, vmt = GD_basic_cdc(self.Gmt, self.Dmt)
        self.assertAlmostEquals(adc, self.adc)
        self.assertAlmostEquals(vdc, self.vdc)
        self.assertAlmostEquals(amt, self.amt)
        self.assertAlmostEquals(vmt, self.vmt)
        a, v = GD_basic_cdc(
            np.array([self.Gmt, self.Gdc]), np.array([self.Dmt, self.Ddc]))
        self.assertAlmostEquals(a[1], self.adc)
        self.assertAlmostEquals(v[1], self.vdc)
        self.assertAlmostEquals(a[0], self.amt)
        self.assertAlmostEquals(v[0], self.vmt)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_TNP_SDR_python(self, logger):
        with PythonOnly():
            [Sdc, Ddc, Rdc] = TNP_SDR(self.Tdc, self.Ndc, self.Pdc)
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            [Smt, Dmt, Rmt] = TNP_SDR(self.Tmt, self.Nmt, self.Pmt)
            try:
                self.assertAlmostEquals(Sdc, self.S1dc)
                self.assertAlmostEquals(Ddc, self.D1dc)
                self.assertAlmostEquals(Rdc, self.R1dc)
            except Exception:
                self.assertAlmostEquals(Sdc, self.S2dc)
                self.assertAlmostEquals(Ddc, self.D2dc)
                self.assertAlmostEquals(Rdc, self.R2dc)
            try:
                self.assertAlmostEquals(Smt, self.S1mt)
                self.assertAlmostEquals(Dmt, self.D1mt)
                self.assertAlmostEquals(Rmt, self.R1mt)
            except Exception:
                self.assertAlmostEquals(Smt, self.S2mt)
                self.assertAlmostEquals(Dmt, self.D2mt)
                self.assertAlmostEquals(Rmt, self.R2mt)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_TNP_SDR_multiple_python(self, logger):
        with PythonOnly():
            [s, d, r] = TNP_SDR(np.append(self.Tmt, self.Tdc, 1),
                                np.append(self.Nmt, self.Ndc, 1),
                                np.append(self.Pmt, self.Pdc, 1))
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            try:
                self.assertAlmostEquals(s[0], self.S1mt)
                self.assertAlmostEquals(d[0], self.D1mt)
                self.assertAlmostEquals(r[0], self.R1mt)
            except Exception:
                self.assertAlmostEquals(s[0], self.S2mt)
                self.assertAlmostEquals(d[0], self.D2mt)
                self.assertAlmostEquals(r[0], self.R2mt)
            try:
                self.assertAlmostEquals(s[1], self.S1dc)
                self.assertAlmostEquals(d[1], self.D1dc)
                self.assertAlmostEquals(r[1], self.R1dc)
            except Exception:
                self.assertAlmostEquals(s[1], self.S2dc)
                self.assertAlmostEquals(d[1], self.D2dc)
                self.assertAlmostEquals(r[1], self.R2dc)
            [S2, D2, R2] = TNP_SDR(self.T2, self.N2, self.P2)
            for i, s in enumerate(S2):
                try:
                    self.assertAlmostEquals(s, self.S21[i])
                    self.assertAlmostEquals(D2[i], self.D21[i])
                    self.assertAlmostEquals(R2[i], self.R21[i])
                except Exception:
                    self.assertAlmostEquals(s, self.S22[i])
                    self.assertAlmostEquals(D2[i], self.D22[i])
                    self.assertAlmostEquals(R2[i], self.R22[i])

            [S2, D2, R2] = TNP_SDR(-self.T2, -self.N2, -self.P2)
            for i, s in enumerate(S2):
                try:
                    self.assertAlmostEquals(s, self.S21[i])
                    self.assertAlmostEquals(D2[i], self.D21[i])
                    self.assertAlmostEquals(R2[i], self.R21[i])
                except Exception:
                    self.assertAlmostEquals(s, self.S22[i])
                    self.assertAlmostEquals(D2[i], self.D22[i])
                    self.assertAlmostEquals(R2[i], self.R22[i])

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_TNP_SDR_cython(self, logger):
        [Sdc, Ddc, Rdc] = TNP_SDR(self.Tdc, self.Ndc, self.Pdc)
        [Smt, Dmt, Rmt] = TNP_SDR(self.Tmt, self.Nmt, self.Pmt)
        logger.info.assert_not_called()
        try:
            self.assertAlmostEquals(Sdc, self.S1dc)
            self.assertAlmostEquals(Ddc, self.D1dc)
            self.assertAlmostEquals(Rdc, self.R1dc)
        except Exception:
            self.assertAlmostEquals(Sdc, self.S2dc)
            self.assertAlmostEquals(Ddc, self.D2dc)
            self.assertAlmostEquals(Rdc, self.R2dc)
        try:
            self.assertAlmostEquals(Smt, self.S1mt)
            self.assertAlmostEquals(Dmt, self.D1mt)
            self.assertAlmostEquals(Rmt, self.R1mt)
        except Exception:
            self.assertAlmostEquals(Smt, self.S2mt)
            self.assertAlmostEquals(Dmt, self.D2mt)
            self.assertAlmostEquals(Rmt, self.R2mt)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_TNP_SDR_multiple_cython(self, logger):
        [s, d, r] = TNP_SDR(np.append(self.Tmt, self.Tdc, 1),
                            np.append(self.Nmt, self.Ndc, 1),
                            np.append(self.Pmt, self.Pdc, 1))
        logger.info.assert_not_called()
        try:
            self.assertAlmostEquals(s[0], self.S1mt)
            self.assertAlmostEquals(d[0], self.D1mt)
            self.assertAlmostEquals(r[0], self.R1mt)
        except Exception:
            self.assertAlmostEquals(s[0], self.S2mt)
            self.assertAlmostEquals(d[0], self.D2mt)
            self.assertAlmostEquals(r[0], self.R2mt)
        try:
            self.assertAlmostEquals(s[1], self.S1dc)
            self.assertAlmostEquals(d[1], self.D1dc)
            self.assertAlmostEquals(r[1], self.R1dc)
        except Exception:
            self.assertAlmostEquals(s[1], self.S2dc)
            self.assertAlmostEquals(d[1], self.D2dc)
            self.assertAlmostEquals(r[1], self.R2dc)
        [S2, D2, R2] = TNP_SDR(self.T2, self.N2, self.P2)
        for i, s in enumerate(S2):
            try:
                self.assertAlmostEquals(s, self.S21[i])
                self.assertAlmostEquals(D2[i], self.D21[i])
                self.assertAlmostEquals(R2[i], self.R21[i])
            except Exception:
                self.assertAlmostEquals(s, self.S22[i])
                self.assertAlmostEquals(D2[i], self.D22[i])
                self.assertAlmostEquals(R2[i], self.R22[i])
        [S2, D2, R2] = TNP_SDR(-self.T2, -self.N2, -self.P2)
        for i, s in enumerate(S2):
            try:
                self.assertAlmostEquals(s, self.S21[i])
                self.assertAlmostEquals(D2[i], self.D21[i])
                self.assertAlmostEquals(R2[i], self.R21[i])
            except Exception:
                self.assertAlmostEquals(s, self.S22[i])
                self.assertAlmostEquals(D2[i], self.D22[i])
                self.assertAlmostEquals(R2[i], self.R22[i])

    def test_TP_FP(self):
        [N1dc, N2dc] = TP_FP(self.Tdc, self.Pdc)
        [N1mt, N2mt] = TP_FP(self.Tmt, self.Pmt)
        self.assertVectorEquals(N1dc, self.N1dc)
        self.assertVectorEquals(N2dc, self.N2dc)
        self.assertVectorEquals(N1mt, self.N1mt)
        self.assertVectorEquals(N2mt, self.N2mt)
        [N1, N2] = TP_FP(
            np.append(self.Tmt, self.Tdc, 1), np.append(self.Pmt, self.Pdc, 1))
        self.assertVectorEquals(N1[:, 0], self.N1mt)
        self.assertVectorEquals(N2[:, 0], self.N2mt)
        self.assertVectorEquals(N1[:, 1], self.N1dc)
        self.assertVectorEquals(N2[:, 1], self.N2dc)

    def test_FP_SDR(self):
        [Sdc, Ddc, Rdc] = FP_SDR(np.array(self.N1dc), np.array(self.N2dc))
        [Smt, Dmt, Rmt] = FP_SDR(np.array(self.N1mt), np.array(self.N2mt))
        self.assertAlmostEquals(Sdc, self.S1dc)
        self.assertAlmostEquals(Ddc, self.D1dc)
        self.assertAlmostEquals(Rdc, self.R1dc)
        self.assertAlmostEquals(Smt, self.S1mt)
        self.assertAlmostEquals(Dmt, self.D1mt)
        self.assertAlmostEquals(Rmt, self.R1mt)

    def test_FP_SDR_matrix(self):
        [Sdc, Ddc, Rdc] = FP_SDR(np.matrix(self.N1dc), np.matrix(self.N2dc))
        [Smt, Dmt, Rmt] = FP_SDR(np.matrix(self.N1mt), np.matrix(self.N2mt))
        self.assertAlmostEquals(Sdc, self.S1dc)
        self.assertAlmostEquals(Ddc, self.D1dc)
        self.assertAlmostEquals(Rdc, self.R1dc)
        self.assertAlmostEquals(Smt, self.S1mt)
        self.assertAlmostEquals(Dmt, self.D1mt)
        self.assertAlmostEquals(Rmt, self.R1mt)

    def test_FP_SDR_multiple(self):
        [s, d, r] = FP_SDR(np.append(self.N1mt, self.N1dc, 1),
                           np.append(self.N2mt, self.N2dc, 1))
        self.assertAlmostEquals(s[0], self.S1mt)
        self.assertAlmostEquals(d[0], self.D1mt)
        self.assertAlmostEquals(r[0], self.R1mt)
        self.assertAlmostEquals(s[1], self.S1dc)
        self.assertAlmostEquals(d[1], self.D1dc)
        self.assertAlmostEquals(r[1], self.R1dc)
        [Sdc, Ddc, Rdc] = FP_SDR(self.N2dc, self.N1dc)
        [Smt, Dmt, Rmt] = FP_SDR(self.N2mt, self.N1mt)
        self.assertAlmostEquals(Sdc, self.S2dc)
        self.assertAlmostEquals(Ddc, self.D2dc)
        self.assertAlmostEquals(Rdc, self.R2dc)
        self.assertAlmostEquals(Smt, self.S2mt)
        self.assertAlmostEquals(Dmt, self.D2mt)
        self.assertAlmostEquals(Rmt, self.R2mt)
        [s, d, r] = FP_SDR(np.append(self.N2mt, self.N2dc, 1),
                           np.append(self.N1mt, self.N1dc, 1))
        self.assertAlmostEquals(s[0], self.S2mt)
        self.assertAlmostEquals(d[0], self.D2mt)
        self.assertAlmostEquals(r[0], self.R2mt)
        self.assertAlmostEquals(s[1], self.S2dc)
        self.assertAlmostEquals(d[1], self.D2dc)
        self.assertAlmostEquals(r[1], self.R2dc)

    def test_FP_SDR_array(self):
        [Sdc, Ddc, Rdc] = FP_SDR(np.array(2*self.N1dc), np.array(2*self.N2dc))
        [Smt, Dmt, Rmt] = FP_SDR(np.array(2*self.N1mt), np.array(2*self.N2mt))
        self.assertAlmostEquals(Sdc, self.S1dc)
        self.assertAlmostEquals(Ddc, self.D1dc)
        self.assertAlmostEquals(Rdc, self.R1dc)
        self.assertAlmostEquals(Smt, self.S1mt)
        self.assertAlmostEquals(Dmt, self.D1mt)
        self.assertAlmostEquals(Rmt, self.R1mt)

    def test_basic_cdc_GD(self):
        Gdc, Ddc = basic_cdc_GD(self.adc, self.vdc)
        Gmt, Dmt = basic_cdc_GD(self.amt, self.vmt)
        self.assertAlmostEquals(Gdc, self.Gdc)
        self.assertAlmostEquals(Ddc, self.Ddc)
        self.assertAlmostEquals(Gmt, self.Gmt)
        self.assertAlmostEquals(Dmt, self.Dmt)
        G, D = basic_cdc_GD(np.array([self.amt, self.adc]),
                            np.array([self.vmt, self.vdc]))
        self.assertAlmostEquals(G[1], self.Gdc)
        self.assertAlmostEquals(D[1], self.Ddc)
        self.assertAlmostEquals(G[0], self.Gmt)
        self.assertAlmostEquals(D[0], self.Dmt)

    def test_basic_cdc_GD_pi_2(self):
        G2, D2 = basic_cdc_GD(np.pi/2, self.vmt)
        self.assertEqual(D2, 0)
        G, D = basic_cdc_GD(np.array([np.pi/2, np.pi/2]),
                            np.array([self.vmt, self.vdc]))
        self.assertAlmostEquals(D[1], 0)
        self.assertAlmostEquals(D[0], 0)

    def test_GD_E(self):
        mtc._CYTHON = False
        Edc = GD_E(self.Gdc, self.Ddc)
        Emt = GD_E(self.Gmt, self.Dmt)
        self.assertAlmostEquals(Edc, self.Edc)
        self.assertAlmostEquals(Emt, self.Emt/np.sqrt(np.diag(self.Emt.T*self.Emt)))
        E = GD_E(np.array([self.Gmt, self.Gdc]),
                 np.array([self.Dmt, self.Ddc]))
        self.assertAlmostEquals(E[:, 1], self.Edc)
        self.assertAlmostEquals(E[:, 0], self.Emt/np.sqrt(np.diag(self.Emt.T*self.Emt)))

    def test_SDR_TNP(self):
        [Tdc, Ndc, Pdc] = SDR_TNP(self.S1dc, self.D1dc, self.R1dc)
        [Tmt, Nmt, Pmt] = SDR_TNP(self.S1mt, self.D1mt, self.R1mt)
        self.assertVectorEquals(Tdc, self.Tdc)
        self.assertVectorEquals(Ndc, self.Ndc)
        self.assertVectorEquals(Pdc, self.Pdc)
        self.assertVectorEquals(Tmt, self.Tmt)
        self.assertVectorEquals(Nmt, self.Nmt)
        self.assertVectorEquals(Pmt, self.Pmt)
        [T, N, P] = SDR_TNP(np.array([self.S1mt, self.S1dc]), np.array(
            [self.D1mt, self.D1dc]), np.array([self.R1mt, self.R1dc]))
        self.assertVectorEquals(T[:, 0], self.Tmt)
        self.assertVectorEquals(N[:, 0], self.Nmt)
        self.assertVectorEquals(P[:, 0], self.Pmt)
        self.assertVectorEquals(T[:, 1], self.Tdc)
        self.assertVectorEquals(N[:, 1], self.Ndc)
        self.assertVectorEquals(P[:, 1], self.Pdc)
        [Tdc, Ndc, Pdc] = SDR_TNP(self.S2dc, self.D2dc, self.R2dc)
        [Tmt, Nmt, Pmt] = SDR_TNP(self.S2mt, self.D2mt, self.R2mt)
        self.assertVectorEquals(Tdc, self.Tdc)
        self.assertVectorEquals(Ndc, self.Ndc)
        self.assertVectorEquals(Pdc, self.Pdc)
        self.assertVectorEquals(Tmt, self.Tmt)
        self.assertVectorEquals(Nmt, self.Nmt)
        self.assertVectorEquals(Pmt, self.Pmt)
        [T, N, P] = SDR_TNP(np.array([self.S2mt, self.S2dc]), np.array(
            [self.D2mt, self.D2dc]), np.array([self.R2mt, self.R2dc]))
        self.assertVectorEquals(T[:, 0], self.Tmt)
        self.assertVectorEquals(N[:, 0], self.Nmt)
        self.assertVectorEquals(P[:, 0], self.Pmt)
        self.assertVectorEquals(T[:, 1], self.Tdc)
        self.assertVectorEquals(N[:, 1], self.Ndc)
        self.assertVectorEquals(P[:, 1], self.Pdc)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_SDR_SDR_python(self, logger):
        with PythonOnly():
            [S2dc, D2dc, R2dc] = SDR_SDR(self.S1dc, self.D1dc, self.R1dc)
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            [S2mt, D2mt, R2mt] = SDR_SDR(self.S1mt, self.D1mt, self.R1mt)
            self.assertAlmostEquals(S2dc, self.S2dc)
            self.assertAlmostEquals(D2dc, self.D2dc)
            self.assertSigmaEquals(R2dc, self.R2dc)
            self.assertAlmostEquals(S2mt, self.S2mt)
            self.assertAlmostEquals(D2mt, self.D2mt)
            self.assertSigmaEquals(R2mt, self.R2mt)
            [S, D, R] = SDR_SDR(np.array([self.S1mt, self.S1dc]), np.array(
                [self.D1mt, self.D1dc]), np.array([self.R1mt, self.R1dc]))
            self.assertAlmostEquals(S[0], self.S2mt)
            self.assertAlmostEquals(D[0], self.D2mt)
            self.assertSigmaEquals(R[0], self.R2mt)
            self.assertAlmostEquals(S[1], self.S2dc)
            self.assertAlmostEquals(D[1], self.D2dc)
            self.assertSigmaEquals(R[1], self.R2dc)
            [S1dc, D1dc, R1dc] = SDR_SDR(self.S2dc, self.D2dc, self.R2dc)
            [S1mt, D1mt, R1mt] = SDR_SDR(self.S2mt, self.D2mt, self.R2mt)
            try:
                self.assertAlmostEquals(S1dc, self.S1dc)
            except AssertionError as e:
                if self.S1dc in [0, np.pi]:
                    self.assertAlmostEquals(S1dc, np.mod(self.S1dc+np.pi, 2*np.pi))
                else:
                    raise e
            self.assertAlmostEquals(D1dc, self.D1dc)
            self.assertSigmaEquals(R1dc, self.R1dc)
            self.assertAlmostEquals(S1mt, self.S1mt)
            self.assertAlmostEquals(D1mt, self.D1mt)
            self.assertSigmaEquals(R1mt, self.R1mt)
            [S, D, R] = SDR_SDR(np.array([self.S2mt, self.S2dc]), np.array(
                [self.D2mt, self.D2dc]), np.array([self.R2mt, self.R2dc]))
            self.assertAlmostEquals(S[0], self.S1mt)
            self.assertAlmostEquals(D[0], self.D1mt)
            self.assertSigmaEquals(R[0], self.R1mt)
            try:
                self.assertAlmostEquals(S[1], self.S1dc)
            except AssertionError as e:
                if self.S1dc in [0, np.pi]:
                    self.assertAlmostEquals(S[1], np.mod(self.S1dc+np.pi, 2*np.pi))
                else:
                    raise e
            self.assertAlmostEquals(D[1], self.D1dc)
            self.assertSigmaEquals(R[1], self.R1dc)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_SDR_SDR_cython(self, logger):
        [S2dc, D2dc, R2dc] = SDR_SDR(self.S1dc, self.D1dc, self.R1dc)
        [S2mt, D2mt, R2mt] = SDR_SDR(self.S1mt, self.D1mt, self.R1mt)
        self.assertAlmostEquals(S2dc, self.S2dc)
        self.assertAlmostEquals(D2dc, self.D2dc)
        self.assertSigmaEquals(R2dc, self.R2dc)
        self.assertAlmostEquals(S2mt, self.S2mt)
        self.assertAlmostEquals(D2mt, self.D2mt)
        self.assertSigmaEquals(R2mt, self.R2mt)
        [S, D, R] = SDR_SDR(np.array([self.S1mt, self.S1dc]), np.array(
            [self.D1mt, self.D1dc]), np.array([self.R1mt, self.R1dc]))
        self.assertAlmostEquals(S[0], self.S2mt)
        self.assertAlmostEquals(D[0], self.D2mt)
        self.assertSigmaEquals(R[0], self.R2mt)
        self.assertAlmostEquals(S[1], self.S2dc)
        self.assertAlmostEquals(D[1], self.D2dc)
        self.assertSigmaEquals(R[1], self.R2dc)
        [S1dc, D1dc, R1dc] = SDR_SDR(self.S2dc, self.D2dc, self.R2dc)
        [S1mt, D1mt, R1mt] = SDR_SDR(self.S2mt, self.D2mt, self.R2mt)
        try:
            self.assertAlmostEquals(S1dc, self.S1dc)
        except AssertionError as e:
            if self.S1dc in [0, np.pi]:
                self.assertAlmostEquals(S1dc, np.mod(self.S1dc+np.pi, 2*np.pi))
            else:
                raise e
        self.assertAlmostEquals(D1dc, self.D1dc)
        self.assertSigmaEquals(R1dc, self.R1dc)
        self.assertAlmostEquals(S1mt, self.S1mt)
        self.assertAlmostEquals(D1mt, self.D1mt)
        self.assertSigmaEquals(R1mt, self.R1mt)
        [S, D, R] = SDR_SDR(np.array([self.S2mt, self.S2dc]), np.array(
            [self.D2mt, self.D2dc]), np.array([self.R2mt, self.R2dc]))
        self.assertAlmostEquals(S[0], self.S1mt)
        self.assertAlmostEquals(D[0], self.D1mt)
        self.assertSigmaEquals(R[0], self.R1mt)
        try:
            self.assertAlmostEquals(S[1], self.S1dc)
        except AssertionError as e:
            if self.S1dc in [0, np.pi]:
                self.assertAlmostEquals(S[1], np.mod(self.S1dc+np.pi, 2*np.pi))
            else:
                raise e
        self.assertAlmostEquals(D[1], self.D1dc)
        self.assertSigmaEquals(R[1], self.R1dc)
        logger.info.assert_not_called()

    def test_FP_TNP(self):
        [Tdc, Ndc, Pdc] = FP_TNP(self.N1dc, self.N2dc)
        [Tmt, Nmt, Pmt] = FP_TNP(self.N1mt, self.N2mt)
        self.assertVectorEquals(Tdc, self.Tdc)
        self.assertVectorEquals(Ndc, self.Ndc)
        self.assertVectorEquals(Pdc, self.Pdc)
        self.assertVectorEquals(Tmt, self.Tmt)
        self.assertVectorEquals(Nmt, self.Nmt)
        self.assertVectorEquals(Pmt, self.Pmt)
        [T, N, P] = FP_TNP(np.append(self.N1mt, self.N1dc, 1),
                           np.append(self.N2mt, self.N2dc, 1))
        self.assertVectorEquals(T[:, 0], self.Tmt)
        self.assertVectorEquals(N[:, 0], self.Nmt)
        self.assertVectorEquals(P[:, 0], self.Pmt)
        self.assertVectorEquals(T[:, 1], self.Tdc)
        self.assertVectorEquals(N[:, 1], self.Ndc)
        self.assertVectorEquals(P[:, 1], self.Pdc)

    def test_SDSD_FP(self):
        mtc._CYTHON = False
        [N1dc, N2dc] = SDSD_FP(self.S1dc, self.D1dc, self.S2dc, self.D2dc)
        [N1mt, N2mt] = SDSD_FP(self.S1mt, self.D1mt, self.S2mt, self.D2mt)
        try:
            self.assertVectorEquals(N1dc, self.N1dc)
            self.assertVectorEquals(N2dc, self.N2dc)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2dc, self.N1dc)
                self.assertVectorEquals(N1dc, self.N2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertVectorEquals(N1mt, self.N1mt)
            self.assertVectorEquals(N2mt, self.N2mt)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2mt, self.N1mt)
                self.assertVectorEquals(N1mt, self.N2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_SDSD_FP_multiple(self):
        [N1, N2] = SDSD_FP(np.array([self.S1mt, self.S1dc]),
                           np.array([self.D1mt, self.D1dc]),
                           np.array([self.S2mt, self.S2dc]),
                           np.array([self.D2mt, self.D2dc]))
        try:
            self.assertVectorEquals(N1[:, 1], self.N1dc)
            self.assertVectorEquals(N2[:, 1], self.N2dc)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2[:, 1], self.N1dc)
                self.assertVectorEquals(N1[:, 1], self.N2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertVectorEquals(N1[:, 0], self.N1mt)
            self.assertVectorEquals(N2[:, 0], self.N2mt)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2[:, 0], self.N1mt)
                self.assertVectorEquals(N1[:, 0], self.N2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_SDR_FP(self):
        [N1dc, N2dc] = SDR_FP(self.S1dc, self.D1dc, self.R1dc)
        [N1mt, N2mt] = SDR_FP(self.S1mt, self.D1mt, self.R1mt)
        try:
            self.assertVectorEquals(N1dc, self.N1dc)
            self.assertVectorEquals(N2dc, self.N2dc)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2dc, self.N1dc)
                self.assertVectorEquals(N1dc, self.N2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertVectorEquals(N1mt, self.N1mt)
            self.assertVectorEquals(N2mt, self.N2mt)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2mt, self.N1mt)
                self.assertVectorEquals(N1mt, self.N2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_SDR_FP_multiple(self):
        [N1, N2] = SDR_FP(np.array([self.S1mt, self.S1dc]),
                          np.array([self.D1mt, self.D1dc]),
                          np.array([self.R1mt, self.R1dc]))
        try:
            self.assertVectorEquals(N1[:, 1], self.N1dc)
            self.assertVectorEquals(N2[:, 1], self.N2dc)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2[:, 1], self.N1dc)
                self.assertVectorEquals(N1[:, 1], self.N2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertVectorEquals(N1[:, 0], self.N1mt)
            self.assertVectorEquals(N2[:, 0], self.N2mt)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2[:, 0], self.N1mt)
                self.assertVectorEquals(N1[:, 0], self.N2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        [N1dc, N2dc] = SDR_FP(self.S2dc, self.D2dc, self.R2dc)
        [N1mt, N2mt] = SDR_FP(self.S2mt, self.D2mt, self.R2mt)
        try:
            self.assertVectorEquals(N1dc, self.N1dc)
            self.assertVectorEquals(N2dc, self.N2dc)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2dc, self.N1dc)
                self.assertVectorEquals(N1dc, self.N2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertVectorEquals(N1mt, self.N1mt)
            self.assertVectorEquals(N2mt, self.N2mt)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2mt, self.N1mt)
                self.assertVectorEquals(N1mt, self.N2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        [N1, N2] = SDR_FP(np.array([self.S2mt, self.S2dc]), np.array(
            [self.D2mt, self.D2dc]), np.array([self.R2mt, self.R2dc]))
        try:
            self.assertVectorEquals(N1[:, 1], self.N1dc)
            self.assertVectorEquals(N2[:, 1], self.N2dc)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2[:, 1], self.N1dc)
                self.assertVectorEquals(N1[:, 1], self.N2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertVectorEquals(N1[:, 0], self.N1mt)
            self.assertVectorEquals(N2[:, 0], self.N2mt)
        except AssertionError as e1:
            try:
                self.assertVectorEquals(N2[:, 0], self.N1mt)
                self.assertVectorEquals(N1[:, 0], self.N2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_SDR_SDSD(self):
        [S1dc, D1dc, S2dc, D2dc] = SDR_SDSD(self.S1dc, self.D1dc, self.R1dc)
        [S1mt, D1mt, S2mt, D2mt] = SDR_SDSD(self.S1mt, self.D1mt, self.R1mt)
        try:
            self.assertAlmostEquals(S1dc, self.S1dc)
            self.assertAlmostEquals(S2dc, self.S2dc)
            self.assertAlmostEquals(D1dc, self.D1dc)
            self.assertAlmostEquals(D2dc, self.D2dc)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2dc, self.S1dc)
                self.assertAlmostEquals(S1dc, self.S2dc)
                self.assertAlmostEquals(D2dc, self.D1dc)
                self.assertAlmostEquals(D1dc, self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1mt, self.S1mt)
            self.assertAlmostEquals(S2mt, self.S2mt)
            self.assertAlmostEquals(D1mt, self.D1mt)
            self.assertAlmostEquals(D2mt, self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2mt, self.S1mt)
                self.assertAlmostEquals(S1mt, self.S2mt)
                self.assertAlmostEquals(D2mt, self.D1mt)
                self.assertAlmostEquals(D1mt, self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_SDR_SDSD_multiple(self):
        [S1, D1, S2, D2] = SDR_SDSD(np.array([self.S1mt, self.S1dc]),
                                    np.array([self.D1mt, self.D1dc]),
                                    np.array([self.R1mt, self.R1dc]))
        try:
            self.assertAlmostEquals(S1[1], self.S1dc)
            self.assertAlmostEquals(S2[1], self.S2dc)
            self.assertAlmostEquals(D1[1], self.D1dc)
            self.assertAlmostEquals(D2[1], self.D2dc)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2[1], self.S1dc)
                self.assertAlmostEquals(S1[1], self.S2dc)
                self.assertAlmostEquals(D2[1], self.D1dc)
                self.assertAlmostEquals(D1[1], self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1[0], self.S1mt)
            self.assertAlmostEquals(S2[0], self.S2mt)
            self.assertAlmostEquals(D1[0], self.D1mt)
            self.assertAlmostEquals(D2[0], self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2[0], self.S1mt)
                self.assertAlmostEquals(S1[0], self.S2mt)
                self.assertAlmostEquals(D2[0], self.D1mt)
                self.assertAlmostEquals(D1[0], self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        [S1dc, D1dc, S2dc, D2dc] = SDR_SDSD(self.S2dc, self.D2dc, self.R2dc)
        [S1mt, D1mt, S2mt, D2mt] = SDR_SDSD(self.S2mt, self.D2mt, self.R2mt)
        try:
            try:
                self.assertAlmostEquals(S1dc, self.S1dc)
            except AssertionError as e:
                if self.S1dc in [0, np.pi]:
                    self.assertAlmostEquals(
                        S1dc, np.mod(self.S1dc+np.pi, 2*np.pi))
                else:
                    raise e
            try:
                self.assertAlmostEquals(S2dc, self.S2dc)
            except AssertionError as e:
                if self.S2dc in [0, np.pi]:
                    self.assertAlmostEquals(
                        S2dc, np.mod(self.S2dc+np.pi, 2*np.pi))
                else:
                    raise e
            self.assertAlmostEquals(D1dc, self.D1dc)
            self.assertAlmostEquals(D2dc, self.D2dc)
        except AssertionError as e1:
            try:
                try:
                    self.assertAlmostEquals(S2dc, self.S1dc)
                except AssertionError as e:
                    if self.S1dc in [0, np.pi]:
                        self.assertAlmostEquals(
                            S2dc, np.mod(self.S1dc+np.pi, 2*np.pi))
                    else:
                        raise e
                try:
                    self.assertAlmostEquals(S1dc, self.S2dc)
                except AssertionError as e:
                    if self.S2dc in [0, np.pi]:
                        self.assertAlmostEquals(
                            S1dc, np.mod(self.S2dc+np.pi, 2*np.pi))
                    else:
                        raise e
                self.assertAlmostEquals(D2dc, self.D1dc)
                self.assertAlmostEquals(D1dc, self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1mt, self.S1mt)
            self.assertAlmostEquals(S2mt, self.S2mt)
            self.assertAlmostEquals(D1mt, self.D1mt)
            self.assertAlmostEquals(D2mt, self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2mt, self.S1mt)
                self.assertAlmostEquals(S1mt, self.S2mt)
                self.assertAlmostEquals(D2mt, self.D1mt)
                self.assertAlmostEquals(D1mt, self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        [S1, D1, S2, D2] = SDR_SDSD(np.array([self.S2mt, self.S2dc]),
                                    np.array([self.D2mt, self.D2dc]),
                                    np.array([self.R2mt, self.R2dc]))
        try:
            try:
                self.assertAlmostEquals(S1[1], self.S1dc)
            except AssertionError as e:
                if self.S1dc in [0, np.pi]:
                    self.assertAlmostEquals(
                        S1[1], np.mod(self.S1dc+np.pi, 2*np.pi))
                else:
                    raise e
            try:
                self.assertAlmostEquals(S2[1], self.S2dc)
            except AssertionError as e:
                if self.S2dc in [0, np.pi]:
                    self.assertAlmostEquals(
                        S2[1], np.mod(self.S2dc+np.pi, 2*np.pi))
                else:
                    raise e
            self.assertAlmostEquals(D1[1], self.D1dc)
            self.assertAlmostEquals(D2[1], self.D2dc)
        except AssertionError as e1:
            try:
                try:
                    self.assertAlmostEquals(S2[1], self.S1dc)
                except AssertionError as e:
                    if self.S1dc in [0, np.pi]:
                        self.assertAlmostEquals(
                            S2[1], np.mod(self.S1dc+np.pi, 2*np.pi))
                    else:
                        raise e
                try:
                    self.assertAlmostEquals(S1[1], self.S2dc)
                except AssertionError as e:
                    if self.S2dc in [0, np.pi]:
                        self.assertAlmostEquals(
                            S1[1], np.mod(self.S2dc+np.pi, 2*np.pi))
                    else:
                        raise e
                self.assertAlmostEquals(D2[1], self.D1dc)
                self.assertAlmostEquals(D1[1], self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1[0], self.S1mt)
            self.assertAlmostEquals(S2[0], self.S2mt)
            self.assertAlmostEquals(D1[0], self.D1mt)
            self.assertAlmostEquals(D2[0], self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2[0], self.S1mt)
                self.assertAlmostEquals(S1[0], self.S2mt)
                self.assertAlmostEquals(D2[0], self.D1mt)
                self.assertAlmostEquals(D1[0], self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_FP_SDSD(self):
        [S1dc, D1dc, S2dc, D2dc] = FP_SDSD(self.N1dc, self.N2dc)
        [S1mt, D1mt, S2mt, D2mt] = FP_SDSD(self.N1mt, self.N2mt)
        try:
            self.assertAlmostEquals(S1dc, self.S1dc)
            self.assertAlmostEquals(S2dc, self.S2dc)
            self.assertAlmostEquals(D1dc, self.D1dc)
            self.assertAlmostEquals(D2dc, self.D2dc)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2dc, self.S1dc)
                self.assertAlmostEquals(S1dc, self.S2dc)
                self.assertAlmostEquals(D2dc, self.D1dc)
                self.assertAlmostEquals(D1dc, self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1mt, self.S1mt)
            self.assertAlmostEquals(S2mt, self.S2mt)
            self.assertAlmostEquals(D1mt, self.D1mt)
            self.assertAlmostEquals(D2mt, self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2mt, self.S1mt)
                self.assertAlmostEquals(S1mt, self.S2mt)
                self.assertAlmostEquals(D2mt, self.D1mt)
                self.assertAlmostEquals(D1mt, self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_FP_SDSD_multiple(self):
        [S1, D1, S2, D2] = FP_SDSD(np.append(self.N1mt, self.N1dc, 1),
                                   np.append(self.N2mt, self.N2dc, 1))
        try:
            self.assertAlmostEquals(S1[1], self.S1dc)
            self.assertAlmostEquals(S2[1], self.S2dc)
            self.assertAlmostEquals(D1[1], self.D1dc)
            self.assertAlmostEquals(D2[1], self.D2dc)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2[1], self.S1dc)
                self.assertAlmostEquals(S1[1], self.S2dc)
                self.assertAlmostEquals(D2[1], self.D1dc)
                self.assertAlmostEquals(D1[1], self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1[0], self.S1mt)
            self.assertAlmostEquals(S2[0], self.S2mt)
            self.assertAlmostEquals(D1[0], self.D1mt)
            self.assertAlmostEquals(D2[0], self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2[0], self.S1mt)
                self.assertAlmostEquals(S1[0], self.S2mt)
                self.assertAlmostEquals(D2[0], self.D1mt)
                self.assertAlmostEquals(D1[0], self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        [S1dc, D1dc, S2dc, D2dc] = FP_SDSD(self.N2dc, self.N1dc)
        [S1mt, D1mt, S2mt, D2mt] = FP_SDSD(self.N2mt, self.N1mt)
        try:
            self.assertAlmostEquals(S1dc, self.S1dc)
            self.assertAlmostEquals(S2dc, self.S2dc)
            self.assertAlmostEquals(D1dc, self.D1dc)
            self.assertAlmostEquals(D2dc, self.D2dc)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2dc, self.S1dc)
                self.assertAlmostEquals(S1dc, self.S2dc)
                self.assertAlmostEquals(D2dc, self.D1dc)
                self.assertAlmostEquals(D1dc, self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1mt, self.S1mt)
            self.assertAlmostEquals(S2mt, self.S2mt)
            self.assertAlmostEquals(D1mt, self.D1mt)
            self.assertAlmostEquals(D2mt, self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2mt, self.S1mt)
                self.assertAlmostEquals(S1mt, self.S2mt)
                self.assertAlmostEquals(D2mt, self.D1mt)
                self.assertAlmostEquals(D1mt, self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        [S1, D1, S2, D2] = FP_SDSD(np.append(self.N2mt, self.N2dc, 1),
                                   np.append(self.N1mt, self.N1dc, 1))
        try:
            self.assertAlmostEquals(S1[1], self.S1dc)
            self.assertAlmostEquals(S2[1], self.S2dc)
            self.assertAlmostEquals(D1[1], self.D1dc)
            self.assertAlmostEquals(D2[1], self.D2dc)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2[1], self.S1dc)
                self.assertAlmostEquals(S1[1], self.S2dc)
                self.assertAlmostEquals(D2[1], self.D1dc)
                self.assertAlmostEquals(D1[1], self.D2dc)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)
        try:
            self.assertAlmostEquals(S1[0], self.S1mt)
            self.assertAlmostEquals(S2[0], self.S2mt)
            self.assertAlmostEquals(D1[0], self.D1mt)
            self.assertAlmostEquals(D2[0], self.D2mt)
        except AssertionError as e1:
            try:
                self.assertAlmostEquals(S2[0], self.S1mt)
                self.assertAlmostEquals(S1[0], self.S2mt)
                self.assertAlmostEquals(D2[0], self.D1mt)
                self.assertAlmostEquals(D1[0], self.D2mt)
            except AssertionError as e2:
                raise AssertionError(e1.message+' or '+e2.message)

    def test_Tape_MT33(self):
        MT33 = Tape_MT33(self.Gmt, self.Dmt, self.Kmt, self.Hmt, self.Omt)
        self.assertAlmostEquals(MT33, self.MT33/self.MT33norm)
        DC33 = Tape_MT33(self.Gdc, self.Ddc, self.Kdc, self.Hdc, self.Odc)
        self.assertAlmostEquals(DC33, self.DC33)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_Tape_MT6_python(self, logger):
        with PythonOnly():
            MT6 = Tape_MT6(self.Gmt, self.Dmt, self.Kmt, self.Hmt, self.Omt)
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            self.assertAlmostEquals(MT6, self.MT6/self.MT6norm)
            DC6 = Tape_MT6(self.Gdc, self.Ddc, self.Kdc, self.Hdc, self.Odc)
            self.assertAlmostEquals(DC6, self.DC6)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_Tape_MT6_multiple_python(self, logger):
        with PythonOnly():
            MTs = Tape_MT6(np.array([self.Gmt, self.Gdc]),
                           np.array([self.Dmt, self.Ddc]),
                           np.array([self.Kmt, self.Kdc]),
                           np.array([self.Hmt, self.Hdc]),
                           np.array([self.Omt, self.Odc]))
            self.assertAlmostEquals(MTs[:, 0], self.MT6/self.MT6norm)
            self.assertAlmostEquals(MTs[:, 1], self.DC6)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_Tape_MT6_cython(self, logger):
        MT6 = Tape_MT6(self.Gmt, self.Dmt, self.Kmt, self.Hmt, self.Omt)
        self.assertAlmostEquals(MT6, self.MT6/self.MT6norm)
        DC6 = Tape_MT6(self.Gdc, self.Ddc, self.Kdc, self.Hdc, self.Odc)
        self.assertAlmostEquals(DC6, self.DC6)
        logger.info.assert_not_called()

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_Tape_MT6_multiple_cython(self, logger):
        MTs = Tape_MT6(np.array([self.Gmt, self.Gdc]),
                       np.array([self.Dmt, self.Ddc]),
                       np.array([self.Kmt, self.Kdc]),
                       np.array([self.Hmt, self.Hdc]),
                       np.array([self.Omt, self.Odc]))
        self.assertAlmostEquals(MTs[:, 0], self.MT6/self.MT6norm)
        self.assertAlmostEquals(MTs[:, 1], self.DC6)
        logger.info.assert_not_called()

    def test_Tape_TNPE(self):
        [Tmt, Nmt, Pmt, Emt] = Tape_TNPE(self.Gmt, self.Dmt, self.Kmt, self.Hmt, self.Omt)
        self.assertVectorEquals(Tmt, self.Tmt, 4)
        self.assertVectorEquals(Nmt, self.Nmt, 4)
        self.assertVectorEquals(Pmt, self.Pmt, 4)
        Emtnorm = np.sqrt(self.Emt.T*self.Emt)
        self.assertAlmostEquals(Emt, self.Emt/Emtnorm, 4)
        [Tdc, Ndc, Pdc, Edc] = Tape_TNPE(self.Gdc, self.Ddc, self.Kdc, self.Hdc, self.Odc)
        self.assertVectorEquals(Tdc, self.Tdc, 4)
        self.assertVectorEquals(Ndc, self.Ndc, 4)
        self.assertVectorEquals(Pdc, self.Pdc, 4)
        self.assertAlmostEquals(Edc, self.Edc, 4)
        [T, N, P, E] = Tape_TNPE(np.array([self.Gmt, self.Gdc]),
                                 np.array([self.Dmt, self.Ddc]),
                                 np.array([self.Kmt, self.Kdc]),
                                 np.array([self.Hmt, self.Hdc]),
                                 np.array([self.Omt, self.Odc]))
        self.assertVectorEquals(T[:, 0], self.Tmt, 4)
        self.assertVectorEquals(N[:, 0], self.Nmt, 4)
        self.assertVectorEquals(P[:, 0], self.Pmt, 4)
        self.assertAlmostEquals(E[:, 0], self.Emt/Emtnorm, 4)
        self.assertVectorEquals(T[:, 1], self.Tdc, 4)
        self.assertVectorEquals(N[:, 1], self.Ndc, 4)
        self.assertVectorEquals(P[:, 1], self.Pdc, 4)
        self.assertAlmostEquals(E[:, 1], self.Edc, 4)

    def test_normal_SD(self):
        s, d = normal_SD(self.N1dc)
        self.assertAlmostEquals(s, self.S1dc)
        self.assertAlmostEquals(d, self.D1dc)
        s, d = normal_SD(self.N1mt)
        self.assertAlmostEquals(s, self.S1mt)
        self.assertAlmostEquals(d, self.D1mt)
        s, d = normal_SD(self.N2dc)
        self.assertAlmostEquals(s, self.S2dc)
        self.assertAlmostEquals(d, self.D2dc)
        s, d = normal_SD(self.N2mt)
        self.assertAlmostEquals(s, self.S2mt)
        self.assertAlmostEquals(d, self.D2mt)

    def test_normal_SD_array(self):
        s, d = normal_SD(np.array(self.N1dc))
        self.assertAlmostEquals(s, self.S1dc)
        self.assertAlmostEquals(d, self.D1dc)

    def test_toa_vec_radians(self):
        vec = toa_vec(np.pi/4, np.pi/4, True)
        self.assertVectorEquals(vec, np.matrix([[0.5], [0.5], [1/np.sqrt(2)]]))

    def test_toa_vec_degrees(self):
        vec = toa_vec(45., 45., False)
        self.assertVectorEquals(vec, np.matrix([[0.5], [0.5], [1/np.sqrt(2)]]))

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_output_convert_mt(self, logger):
        with PythonOnly():
            result_mt = output_convert(self.MT6)
            self.assertEqual(logger.info.call_args_list[0], mock.call(C_EXTENSION_FALLBACK_LOG_MSG))
            for key in result_mt.keys():
                new_key = key
                if key == 's':
                    new_key = 'O'
                if key in ['S1', 'D1', 'S2', 'D2']:
                    try:
                        self.assertAlmostEquals(result_mt[key],
                                                getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                    except AssertionError as e:
                        if key in ['S1', 'D1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertAlmostEquals(result_mt[key],
                                                    getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                        except AssertionError as e2:
                            try:
                                self.assertAlmostEquals(result_mt[key]-180,
                                                        getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                            except AssertionError as e3:
                                raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

                elif key in ['R1', 'R2']:
                    try:
                        self.assertSigmaEquals(result_mt[key],
                                               getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                    except AssertionError as e:
                        if key in ['R1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertSigmaEquals(result_mt[key],
                                                   getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                        except AssertionError as e2:
                            raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)

                elif key in ['k']:
                    try:
                        self.assertSigmaEquals(result_mt[key],
                                               getattr(self, new_key.upper()+'mt'), 5)
                    except AssertionError as e:
                        try:
                            self.assertSigmaEquals(result_mt[key]-np.pi,
                                                   getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e2:
                            raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)
                else:
                    try:
                        self.assertAlmostEquals(result_mt[key],
                                                getattr(self, new_key.upper()+'mt'), 5)
                    except AssertionError as e:
                        raise AssertionError('Key:'+str(key)+'\n'+e.message)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_output_convert_dc(self, logger):
        with PythonOnly():
            result_dc = output_convert(self.DC6)
            self.assertEqual(logger.info.call_args_list[0], mock.call(C_EXTENSION_FALLBACK_LOG_MSG))
            for key in result_dc.keys():
                new_key = key
                if key == 's':
                    new_key = 'O'
                if key in ['S1', 'D1', 'S2', 'D2']:
                    try:
                        self.assertAlmostEquals(result_dc[key],
                                                getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                    except AssertionError as e:
                        if key in ['S1', 'D1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertAlmostEquals(result_dc[key],
                                                    getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                        except AssertionError as e2:
                            try:
                                self.assertAlmostEquals(result_dc[key]-180, getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                            except AssertionError as e3:
                                raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

                elif key in ['R1', 'R2']:
                    try:
                        self.assertSigmaEquals(result_dc[key]*np.pi/180., getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e:
                        if key in ['R1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertSigmaEquals(result_dc[key]*np.pi/180., getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e2:
                            raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)

                elif key in ['k']:
                    try:
                        self.assertSigmaEquals(result_dc[key],
                                               getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e:
                        try:
                            self.assertSigmaEquals(result_dc[key]-np.pi,
                                                   getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e2:
                            raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)
                else:
                    try:
                        self.assertAlmostEquals(result_dc[key],
                                                getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e:
                        raise AssertionError('Key:'+str(key)+'\n'+e.message)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_output_convert_multiple(self, logger):
        with PythonOnly():
            result = output_convert(np.append(self.MT6, self.DC6, 1))
            self.assertEqual(logger.info.call_args_list[0], mock.call(C_EXTENSION_FALLBACK_LOG_MSG))
            for key in result.keys():
                new_key = key
                if key == 's':
                    new_key = 'O'
                try:
                    self.assertAlmostEquals(result[key][:, 0],
                                            getattr(self, new_key.upper()+'mt'), 5)
                except IndexError:
                    if key in ['S1', 'D1', 'S2', 'D2']:
                        try:
                            self.assertAlmostEquals(result[key][0]*np.pi/180,
                                                    getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e:
                            if key in ['S1', 'D1']:
                                new_key = key.replace('1', '2')
                            else:
                                new_key = key.replace('2', '1')
                            try:
                                self.assertAlmostEquals(result[key][0]*np.pi/180,
                                                        getattr(self, new_key.upper()+'mt'), 5)
                            except AssertionError as e2:
                                try:
                                    self.assertAlmostEquals(result[key][0]-180,
                                                            getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                                except AssertionError as e3:
                                    raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

                    elif key in ['R1', 'R2']:
                        try:
                            self.assertSigmaEquals(result[key][0]*np.pi/180,
                                                   getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e:
                            if key in ['R1']:
                                new_key = key.replace('1', '2')
                            else:
                                new_key = key.replace('2', '1')
                            try:
                                self.assertSigmaEquals(result[key][0]*np.pi/180,
                                                       getattr(self, new_key.upper()+'mt'), 5)
                            except AssertionError as e2:
                                raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)

                    elif key in ['k']:
                        try:
                            self.assertSigmaEquals(result[key][0],
                                                   getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e:
                            try:
                                self.assertSigmaEquals(result[key][0]-np.pi,
                                                       getattr(self, new_key.upper()+'mt'), 5)
                            except AssertionError as e2:
                                raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)
                    else:
                        try:
                            self.assertAlmostEquals(result[key][0],
                                                    getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e:
                            raise AssertionError('Key:'+str(key)+'\n'+e.message)
                except AssertionError as e:
                    raise AssertionError('Key:'+str(key)+'\n'+e.message)
                try:
                    self.assertAlmostEquals(result[key][:, 1],
                                            getattr(self, new_key.upper()+'dc'), 5)
                except IndexError:
                    if key in ['S1', 'D1', 'S2', 'D2']:
                        try:
                            self.assertAlmostEquals(result[key][1]*np.pi/180,
                                                    getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e:
                            if key in ['S1', 'D1']:
                                new_key = key.replace('1', '2')
                            else:
                                new_key = key.replace('2', '1')
                            try:
                                self.assertAlmostEquals(result[key][1]*np.pi/180,
                                                        getattr(self, new_key.upper()+'dc'), 5)
                            except AssertionError as e2:
                                try:
                                    self.assertAlmostEquals(result[key][1]-180,
                                                            getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                                except AssertionError as e3:
                                    raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

                    elif key in ['R1', 'R2']:
                        try:
                            self.assertSigmaEquals(result[key][1]*np.pi/180,
                                                   getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e:
                            if key in ['R1']:
                                new_key = key.replace('1', '2')
                            else:
                                new_key = key.replace('2', '1')
                            try:
                                self.assertSigmaEquals(result[key][1]*np.pi/180,
                                                       getattr(self, new_key.upper()+'dc'), 5)
                            except AssertionError as e2:
                                raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)

                    elif key in ['k']:
                        try:
                            self.assertSigmaEquals(result[key][1],
                                                   getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e:
                            try:
                                self.assertSigmaEquals(result[key][1]-np.pi,
                                                       getattr(self, new_key.upper()+'dc'), 5)
                            except AssertionError as e2:
                                raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)
                    else:
                        try:
                            self.assertAlmostEquals(result[key][1],
                                                    getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e:
                            raise AssertionError('Key:'+str(key)+'\n'+e.message)
                except AssertionError as e:
                    raise AssertionError('Key:'+str(key)+'\n'+e.message)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_output_convert_cython(self, logger):
        result_mt = output_convert(self.MT6)
        logger.info.assert_not_called()
        for key in result_mt.keys():
            new_key = key
            if key == 's':
                new_key = 'O'
            if key in ['S1', 'D1', 'S2', 'D2']:
                try:
                    self.assertAlmostEquals(result_mt[key],
                                            getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                except AssertionError as e:
                    if key in ['S1', 'D1']:
                        new_key = key.replace('1', '2')
                    else:
                        new_key = key.replace('2', '1')
                    try:
                        self.assertAlmostEquals(result_mt[key],
                                                getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                    except AssertionError as e2:
                        try:
                            self.assertAlmostEquals(result_mt[key]-180,
                                                    getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                        except AssertionError as e3:
                            raise AssertionError(
                                'Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

            elif key in ['R1', 'R2']:
                try:
                    self.assertSigmaEquals(result_mt[key],
                                           getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                except AssertionError as e:
                    if key in ['R1']:
                        new_key = key.replace('1', '2')
                    else:
                        new_key = key.replace('2', '1')
                    try:
                        self.assertSigmaEquals(result_mt[key],
                                               getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                    except AssertionError as e2:
                        raise AssertionError(
                            'Key:'+str(key)+'\n'+e.message+' or '+e2.message)

            elif key in ['k']:
                try:
                    self.assertSigmaEquals(result_mt[key],
                                           getattr(self, new_key.upper()+'mt'), 5)
                except AssertionError as e:
                    try:
                        self.assertSigmaEquals(result_mt[key]-np.pi,
                                               getattr(self, new_key.upper()+'mt'), 5)
                    except AssertionError as e2:
                        raise AssertionError(
                            'Key:'+str(key)+'\n'+e.message+' or '+e2.message)
            else:
                try:
                    self.assertAlmostEquals(result_mt[key],
                                            getattr(self, new_key.upper()+'mt'), 5)
                except AssertionError as e:
                    raise AssertionError('Key:'+str(key)+'\n'+e.message)
        result_dc = output_convert(self.DC6)
        for key in result_dc.keys():
            new_key = key
            if key == 's':
                new_key = 'O'
            if key in ['S1', 'D1', 'S2', 'D2']:
                try:
                    self.assertAlmostEquals(result_dc[key],
                                            getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                except AssertionError as e:
                    if key in ['S1', 'D1']:
                        new_key = key.replace('1', '2')
                    else:
                        new_key = key.replace('2', '1')
                    try:
                        self.assertAlmostEquals(result_dc[key],
                                                getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                    except AssertionError as e2:
                        try:
                            self.assertAlmostEquals(result_dc[key]-180,
                                                    getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                        except AssertionError as e3:
                            raise AssertionError(
                                'Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

            elif key in ['R1', 'R2']:
                try:
                    self.assertSigmaEquals(result_dc[key]*np.pi/180.,
                                           getattr(self, new_key.upper()+'dc'), 5)
                except AssertionError as e:
                    if key in ['R1']:
                        new_key = key.replace('1', '2')
                    else:
                        new_key = key.replace('2', '1')
                    try:
                        self.assertSigmaEquals(result_dc[key]*np.pi/180.,
                                               getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e2:
                        raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)

            elif key in ['k']:
                try:
                    self.assertSigmaEquals(result_dc[key],
                                           getattr(self, new_key.upper()+'dc'), 5)
                except AssertionError as e:
                    try:
                        self.assertSigmaEquals(result_dc[key]-np.pi,
                                               getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e2:
                        raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)
            else:
                try:
                    self.assertAlmostEquals(result_dc[key],
                                            getattr(self, new_key.upper()+'dc'), 5)
                except AssertionError as e:
                    raise AssertionError('Key:'+str(key)+'\n'+e.message)
        result = output_convert(np.append(self.MT6, self.DC6, 1))
        for key in result.keys():
            new_key = key
            if key == 's':
                new_key = 'O'
            try:
                self.assertAlmostEquals(result[key][:, 0],
                                        getattr(self, new_key.upper()+'mt'), 5)
            except IndexError:
                if key in ['S1', 'D1', 'S2', 'D2']:
                    try:
                        self.assertAlmostEquals(result[key][0]*np.pi/180,
                                                getattr(self, new_key.upper()+'mt'), 5)
                    except AssertionError as e:
                        if key in ['S1', 'D1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertAlmostEquals(result[key][0]*np.pi/180,
                                                    getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e2:
                            try:
                                self.assertAlmostEquals(result[key][0]-180,
                                                        getattr(self, new_key.upper()+'mt')*180/np.pi, 5)
                            except AssertionError as e3:
                                raise AssertionError(
                                    'Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

                elif key in ['R1', 'R2']:
                    try:
                        self.assertSigmaEquals(result[key][0]*np.pi/180,
                                               getattr(self, new_key.upper()+'mt'), 5)
                    except AssertionError as e:
                        if key in ['R1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertSigmaEquals(result[key][0]*np.pi/180,
                                                   getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e2:
                            raise AssertionError(
                                'Key:'+str(key)+'\n'+e.message+' or '+e2.message)

                elif key in ['k']:
                    try:
                        self.assertSigmaEquals(result[key][0],
                                               getattr(self, new_key.upper()+'mt'), 5)
                    except AssertionError as e:
                        try:
                            self.assertSigmaEquals(result[key][0]-np.pi,
                                                   getattr(self, new_key.upper()+'mt'), 5)
                        except AssertionError as e2:
                            raise AssertionError('Key:'+str(key)+'\n'+e.message+' or '+e2.message)
                else:
                    try:
                        self.assertAlmostEquals(result[key][0],
                                                getattr(self, new_key.upper()+'mt'), 5)
                    except AssertionError as e:
                        raise AssertionError('Key:'+str(key)+'\n'+e.message)
            except AssertionError as e:
                raise AssertionError('Key:'+str(key)+'\n'+e.message)
            try:
                self.assertAlmostEquals(result[key][:, 1],
                                        getattr(self, new_key.upper()+'dc'), 5)
            except IndexError:
                if key in ['S1', 'D1', 'S2', 'D2']:
                    try:
                        self.assertAlmostEquals(result[key][1]*np.pi/180,
                                                getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e:
                        if key in ['S1', 'D1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertAlmostEquals(result[key][1]*np.pi/180,
                                                    getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e2:
                            try:
                                self.assertAlmostEquals(result[key][1]-180,
                                                        getattr(self, new_key.upper()+'dc')*180/np.pi, 5)
                            except AssertionError as e3:
                                raise AssertionError(
                                    'Key:'+str(key)+'\n'+e.message+' or '+e2.message+' or '+e3.message)

                elif key in ['R1', 'R2']:
                    try:
                        self.assertSigmaEquals(result[key][1]*np.pi/180,
                                               getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e:
                        if key in ['R1']:
                            new_key = key.replace('1', '2')
                        else:
                            new_key = key.replace('2', '1')
                        try:
                            self.assertSigmaEquals(result[key][1]*np.pi/180,
                                                   getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e2:
                            raise AssertionError(
                                'Key:'+str(key)+'\n'+e.message+' or '+e2.message)

                elif key in ['k']:
                    try:
                        self.assertSigmaEquals(result[key][1],
                                               getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e:
                        try:
                            self.assertSigmaEquals(result[key][1]-np.pi,
                                                   getattr(self, new_key.upper()+'dc'), 5)
                        except AssertionError as e2:
                            raise AssertionError(
                                'Key:'+str(key)+'\n'+e.message+' or '+e2.message)
                else:
                    try:
                        self.assertAlmostEquals(result[key][1],
                                                getattr(self, new_key.upper()+'dc'), 5)
                    except AssertionError as e:
                        raise AssertionError('Key:'+str(key)+'\n'+e.message)
            except AssertionError as e:
                raise AssertionError('Key:'+str(key)+'\n'+e.message)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_MT6_biaxes_python(self, logger):
        with PythonOnly():
            # Tested against MATLAB code
            phi, explosion, area_displacement = MT6_biaxes(np.array([2., 1, 0, -1, 0, 0]))
            logger.info.assert_called_with(C_EXTENSION_FALLBACK_LOG_MSG)
            self.assertAlmostEquals(phi, np.array([[0.75983569,  0.75983569],
                                                   [-0.39331989, -0.39331989],
                                                   [0.51763809, -0.51763809]]))
            self.assertAlmostEquals(area_displacement, 1.18301270189222)
            self.assertAlmostEquals(explosion, 0.0849364905389036)
            phi, explosion, area_displacement = MT6_biaxes(
                np.array([[2, 3], [1, 0], [0, -3], [-1, 0], [0, 0], [0, 0]]))
            self.assertAlmostEquals(phi, np.array([[[0.75983569, 0.75983569], [-0.39331989, -0.39331989],
                                                    [0.51763809, -0.51763809]],
                                                   [[0.70710678, 0.70710678], [0., 0.],
                                                    [0.70710678, -0.70710678]]]))
            self.assertAlmostEquals(area_displacement, np.array([1.1830127, 3.]))
            self.assertAlmostEquals(explosion, np.array([0.08493649, 0.]))
            vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                    0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
            phi, explosion, area_displacement = MT6_biaxes(
                np.array([2, 1, 0, -1, 0, 0]), vtic)
            try:
                self.assertVectorEquals(phi[:, 0], np.array([-0.062501878200605, 0.712609970150016, 0.698770738986823]))
                self.assertVectorEquals(phi[:, 1], np.array(
                    [-0.062501878200605, 0.712609970150016, -0.698770738986823]))
            except AssertionError:
                self.assertVectorEquals(phi[:, 1], np.array([-0.062501878200605, 0.712609970150016, 0.698770738986823]))
                self.assertVectorEquals(phi[:, 0], np.array([-0.062501878200605, 0.712609970150016, -0.698770738986823]))
            self.assertAlmostEquals(area_displacement, 1.984495199420505)
            self.assertAlmostEquals(explosion, 0.461237998551262)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_MT6_biaxes_cython(self, logger):
        mtc._CYTHON = True
        # Tested against MATLAB code
        phi, explosion, area_displacement = MT6_biaxes(
            np.array([2, 1, 0, -1, 0, 0]))
        self.assertAlmostEquals(phi, np.array(
            [[0.75983569,  0.75983569], [-0.39331989, -0.39331989], [0.51763809, -0.51763809]]))
        self.assertAlmostEquals(area_displacement, 1.18301270189222)
        self.assertAlmostEquals(explosion, 0.0849364905389036)
        phi, explosion, area_displacement = MT6_biaxes(
            np.array([[2, 3], [1, 0], [0, -3], [-1, 0], [0, 0], [0, 0]]))
        self.assertAlmostEquals(phi, np.array([[[0.75983569, 0.75983569], [-0.39331989, -0.39331989], [0.51763809, -0.51763809]],
                                               [[0.70710678, 0.70710678], [0., 0.], [0.70710678, -0.70710678]]]))
        self.assertAlmostEquals(area_displacement, np.array([1.1830127, 3.]))
        self.assertAlmostEquals(explosion, np.array([0.08493649, 0.]))
        vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
        phi, explosion, area_displacement = MT6_biaxes(
            np.array([2, 1, 0, -1, 0, 0]), vtic)
        try:
            self.assertVectorEquals(
                phi[:, 0], np.array([-0.062501878200605, 0.712609970150016, 0.698770738986823]))
            self.assertVectorEquals(phi[:, 1], np.array(
                [-0.062501878200605, 0.712609970150016, -0.698770738986823]))
        except AssertionError:
            self.assertVectorEquals(
                phi[:, 1], np.array([-0.062501878200605, 0.712609970150016, 0.698770738986823]))
            self.assertVectorEquals(phi[:, 0], np.array(
                [-0.062501878200605, 0.712609970150016, -0.698770738986823]))
        self.assertAlmostEquals(area_displacement, 1.984495199420505)
        self.assertAlmostEquals(explosion, 0.461237998551262)
        logger.info.assert_not_called()

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_MT6c_D6_python(self, logger):
        with PythonOnly():
            D6 = MT6c_D6(self.MT6, isotropic_c())
            logger.info.assert_called_with(C_EXTENSION_FALLBACK_LOG_MSG)
            Td, Nd, Pd, Ed = MT6_TNPE(D6)
            try:
                self.assertAlmostEquals(Td, self.Tmt)
            except Exception:
                self.assertAlmostEquals(Td, -self.Tmt)
            try:
                self.assertAlmostEquals(Nd, self.Nmt)
            except Exception:
                self.assertAlmostEquals(Nd, -self.Nmt)
            try:
                self.assertAlmostEquals(Pd, self.Pmt)
            except Exception:
                self.assertAlmostEquals(Pd, -self.Pmt)
            with self.assertRaises(AssertionError):
                self.assertAlmostEquals(Ed, self.Emt)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_MT6c_D6_cython(self, logger):
        D6 = MT6c_D6(self.MT6, isotropic_c())
        logger.info.assert_not_called()
        Td, Nd, Pd, Ed = MT6_TNPE(D6)
        try:
            self.assertAlmostEquals(Td, self.Tmt)
        except Exception:
            self.assertAlmostEquals(Td, -self.Tmt)
        try:
            self.assertAlmostEquals(Nd, self.Nmt)
        except Exception:
            self.assertAlmostEquals(Nd, -self.Nmt)
        try:
            self.assertAlmostEquals(Pd, self.Pmt)
        except Exception:
            self.assertAlmostEquals(Pd, -self.Pmt)
        with self.assertRaises(AssertionError):
            self.assertAlmostEquals(Ed, self.Emt)

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_c_norm_python(self, logger):
        with PythonOnly():
            self.assertEqual(c_norm(isotropic_c()), 6.7082039324993694)
            self.assertEqual(c_norm(isotropic_c(2)), 9.16515138991168)
            self.assertEqual(c_norm(isotropic_c(2, 4)), 22.715633383201094)
            logger.info.assert_called_with(C_EXTENSION_FALLBACK_LOG_MSG)
            # Matlab values

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_c_norm_cython(self, logger):
        self.assertEqual(c_norm(isotropic_c()), 6.7082039324993694)
        self.assertEqual(c_norm(isotropic_c(2)), 9.16515138991168)
        self.assertEqual(c_norm(isotropic_c(2, 4)), 22.715633383201094)
        # Matlab values
        logger.info.assert_not_called()

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_isotropic_c_python(self, logger):
        with PythonOnly():
            self.assertEqual(isotropic_c(),
                             [3, 1, 1, 0, 0, 0, 3, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 1, 0, 1])
            self.assertEqual(isotropic_c(2),
                             [4, 2, 2, 0, 0, 0, 4, 2, 0, 0, 0, 4, 0, 0, 0, 1, 0, 0, 1, 0, 1])
            vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                    0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
            self.assertAlmostEquals(np.array(isotropic_c(c=vtic)),
                                    np.array([5.2666666666666675, -1.1333333333333333,
                                              -1.1333333333333333, 0, 0, 0,
                                              5.2666666666666675, -1.1333333333333333,
                                              0, 0, 0, 5.2666666666666675, 0, 0, 0,
                                              3.2000000000000002, 0, 0, 3.2000000000000002,
                                              0, 3.2000000000000002]))
            logger.info.assert_called_with(C_EXTENSION_FALLBACK_LOG_MSG)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_isotropic_c_cython(self, logger):
        self.assertEqual(isotropic_c(),
                         [3, 1, 1, 0, 0, 0, 3, 1, 0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 1, 0, 1])
        self.assertEqual(isotropic_c(2),
                         [4, 2, 2, 0, 0, 0, 4, 2, 0, 0, 0, 4, 0, 0, 0, 1, 0, 0, 1, 0, 1])
        vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
        self.assertAlmostEquals(np.array(isotropic_c(c=vtic)),
                                np.array([5.2666666666666675, -1.1333333333333333,
                                          -1.1333333333333333, 0, 0, 0, 5.2666666666666675,
                                          -1.1333333333333333, 0, 0, 0, 5.2666666666666675,
                                          0, 0, 0, 3.2000000000000002, 0, 0, 3.2000000000000002,
                                          0, 3.2000000000000002]))
        logger.info.assert_not_called()

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_is_isotropic_c_python(self, logger):
        with PythonOnly():
            self.assertTrue(is_isotropic_c(isotropic_c()))
            self.assertTrue(is_isotropic_c(isotropic_c(2)))
            vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                    0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
            self.assertFalse(is_isotropic_c(vtic))
            self.assertTrue(is_isotropic_c(isotropic_c(c=vtic)))
            logger.info.assert_called_with(C_EXTENSION_FALLBACK_LOG_MSG)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_is_isotropic_c_cython(self, logger):
        self.assertTrue(is_isotropic_c(isotropic_c()))
        self.assertTrue(is_isotropic_c(isotropic_c(2)))
        vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
        self.assertFalse(is_isotropic_c(vtic))
        self.assertTrue(is_isotropic_c(isotropic_c(c=vtic)))
        logger.info.assert_not_called()

    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_c21_cvoigt_python(self, logger):
        with PythonOnly():
            self.assertAlmostEquals(c21_cvoigt(isotropic_c()), np.array([[3, 1, 1, 0, 0, 0],
                                                                         [1, 3, 1,
                                                                             0, 0, 0],
                                                                         [1, 1, 3,
                                                                             0, 0, 0],
                                                                         [0, 0, 0,
                                                                             2, 0, 0],
                                                                         [0, 0, 0,
                                                                             0, 2, 0],
                                                                         [0, 0, 0, 0, 0, 2]]))
            self.assertAlmostEquals(c21_cvoigt(isotropic_c(2)), np.array([[4, 2, 2, 0, 0, 0],
                                                                          [2, 4, 2,
                                                                              0, 0, 0],
                                                                          [2, 2, 4,
                                                                              0, 0, 0],
                                                                          [0, 0, 0,
                                                                              2, 0, 0],
                                                                          [0, 0, 0,
                                                                              0, 2, 0],
                                                                          [0, 0, 0, 0, 0, 2]]))
            vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                    0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
            self.assertAlmostEquals(c21_cvoigt(vtic), np.array([[1, 2, 0.5, 0, 0, 0],
                                                                [2, 1, 0.5,
                                                                    0, 0, 0],
                                                                [0.5, 0.5, 1,
                                                                    0, 0, 0],
                                                                [0, 0,  0,
                                                                    12, 0, 0],
                                                                [0, 0,  0, 0,
                                                                    12, 0],
                                                                [0, 0,  0, 0, 0, 8]]))
            logger.info.assert_called_with(C_EXTENSION_FALLBACK_LOG_MSG)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.convert.moment_tensor_conversion.logger')
    def test_c21_cvoigt_cython(self, logger):
        self.assertAlmostEquals(c21_cvoigt(isotropic_c()), np.array([[3, 1, 1, 0, 0, 0],
                                                                     [1, 3, 1,
                                                                         0, 0, 0],
                                                                     [1, 1, 3,
                                                                         0, 0, 0],
                                                                     [0, 0, 0,
                                                                         2, 0, 0],
                                                                     [0, 0, 0,
                                                                         0, 2, 0],
                                                                     [0, 0, 0, 0, 0, 2]]))
        self.assertAlmostEquals(c21_cvoigt(isotropic_c(2)), np.array([[4, 2, 2, 0, 0, 0],
                                                                      [2, 4, 2,
                                                                          0, 0, 0],
                                                                      [2, 2, 4,
                                                                          0, 0, 0],
                                                                      [0, 0, 0,
                                                                          2, 0, 0],
                                                                      [0, 0, 0,
                                                                          0, 2, 0],
                                                                      [0, 0, 0, 0, 0, 2]]))
        vtic = [1.0000, 2.0000, 0.5000, 0, 0, 0, 1.0000, 0.5000,
                0, 0, 0, 1.0000, 0, 0, 0, 6.0000, 0, 0, 6.0000, 0, 4.]
        self.assertAlmostEquals(c21_cvoigt(vtic), np.array([[1, 2, 0.5, 0, 0, 0],
                                                            [2, 1, 0.5,
                                                                0, 0, 0],
                                                            [0.5, 0.5, 1,
                                                                0, 0, 0],
                                                            [0, 0,  0,
                                                                12, 0, 0],
                                                            [0, 0,  0, 0,
                                                                12, 0],
                                                            [0, 0,  0, 0, 0, 8]]))
        logger.info.assert_not_called()



import unittest
import os

import numpy as np

from MTfit.utilities.file_io import pickle_output
from MTfit.plot.core import run
from MTfit.plot.core import MTplot

VERBOSITY = 2

PLOT_DIR = os.environ.get('MTFIT_TEST_PLOT_DIR', None)


class CoreTestCase(unittest.TestCase):

    def setUp(self):
        self.multiMTs = np.array([[1, 2, 0, 0, 1, 0, 2, -1],
                                  [0, 0, 0, 0, 0, 0, 1, 0],
                                  [-1, -1, 0, 0, -1, 0, 1, 3],
                                  [0, 0, 1, 0, 0, 1, 0, 1],
                                  [0, 0, 0, 0, 0, 0, 0, 0],
                                  [0, 0, 0, 1, 0, 0, 0, 0]])
        self.multiDCs = np.array([[1, 0, 0, 0],
                                  [0, 1, 0, 0],
                                  [-1, -1, 0, 0],
                                  [0, 0, 1, 0],
                                  [0, 0, 0, 0],
                                  [0, 0, 0, 1]])
        self.singleMT = np.array([1, 2, -2, 0, 1, 0])
        self.station_distribution = {'distribution': [
            {'azimuth': np.matrix([[231.1], [42.9], [21.2], [256.4], [197.4],
                                   [229.7], [75.6], [187.5], [85.3], [307.5], [355.8], [14.7], [
                                       123.5], [231.8], [45.9], [193.8], [53.7], [218.4], [12.9],
                                   [325.5], [29.4], [278.9], [326.1], [223.7], [42.6], [253.6]]),
             'names': ['S0271', 'S0649', 'S0484', 'S0263', 'S0142', 'S0244', 'S0415', 'S0065', 'S0362', 'S0450', 'S0534', 'S0641', 'S0155', 'S0162', 'S0650',
                       'S0195', 'S0517', 'S0004', 'S0588', 'S0377', 'S0618', 'S0347', 'S0529', 'S0083', 'S0595', 'S0236'],
             'takeoff_angle': np.matrix([[154.7], [109.7], [145.4], [122.7], [137.6], [148.1], [122.8], [126.1], [128.2], [137.7], [138.2], [120.2],
                                         [117.], [127.5], [108.2], [147.3], [124.2], [109.8], [128.6], [165.3], [120.5], [149.5], [131.7], [118.2], [117.8], [118.6]])},
            {'azimuth': np.matrix([[230.9], [43.], [21.3], [256.4], [197.3], [229.6], [75.7], [187.4], [85.3], [307.5], [355.7], [14.8],
                                   [123.5], [231.7], [45.9], [193.6], [53.7], [218.3], [13.], [325.7], [29.5], [278.9], [326.1], [223.7], [42.7], [253.5]]),
                'names': ['S0271', 'S0649', 'S0484', 'S0263', 'S0142', 'S0244', 'S0415', 'S0065', 'S0362', 'S0450', 'S0534', 'S0641', 'S0155', 'S0162', 'S0650',
                          'S0195', 'S0517', 'S0004', 'S0588', 'S0377', 'S0618', 'S0347', 'S0529', 'S0083', 'S0595', 'S0236'],
                'takeoff_angle': np.matrix([[154.8], [109.8], [145.4], [122.8], [137.6], [148.1], [122.8], [126.1], [128.2], [137.8], [138.3], [120.3],
                                            [117.1], [127.6], [108.3], [147.3], [124.2], [109.9], [128.7], [165.4], [120.5], [149.6], [131.8], [118.2], [117.9], [118.7]])}],
            'probability': [504.7, 504.7]}
        self.stations = {'azimuth': np.matrix([[231.1], [42.9], [21.2], [256.4], [197.4],
                                               [229.7], [75.6], [187.5], [85.3], [307.5], [355.8], [14.7],
                                               [123.5], [231.8], [45.9], [193.8], [53.7], [218.4], [12.9],
                                               [325.5], [29.4], [278.9], [326.1], [223.7], [42.6], [253.6]]),
                         'names': ['S0271', 'S0649', 'S0484', 'S0263', 'S0142', 'S0244', 'S0415', 'S0065', 'S0362', 'S0450', 'S0534', 'S0641', 'S0155', 'S0162', 'S0650',
                                   'S0195', 'S0517', 'S0004', 'S0588', 'S0377', 'S0618', 'S0347', 'S0529', 'S0083', 'S0595', 'S0236'],
                         'takeoff_angle': np.matrix([[154.7], [109.7], [145.4], [122.7], [137.6], [148.1], [122.8], [126.1], [128.2], [137.7], [138.2], [120.2],
                                                     [117.], [127.5], [108.2], [147.3], [124.2], [109.8], [128.6], [165.3], [120.5], [149.5], [131.7], [118.2], [117.8], [118.6]]),
                         'polarity': [1, 0, 1, 0, -1, 1, -1, 0, 1, 1, -1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 1, -1, 1, 1, -1]}

    def generate_test_file(self):
        # Breaks when MATLAB - not clear why
        pickle_output([{'MTSpace': np.matrix([[1.], [2.], [1.], [2.], [1.], [2.]])}, {}], 'plottest.out')

    def test_run(self):
        self.generate_test_file()
        save_file, show = self.make_savefile('Run Beachball')
        if show:
            opt = []
        else:
            opt = ['-q', '--save_file='+save_file]
        run(['-d=plottest.out']+opt)
        run(['-p=hudson', '-d=plottest.out']+opt)
        run(['-p=lune', '-d=plottest.out']+opt)
        run(['-p=faultplane', '-d=plottest.out']+opt)
        try:
            os.remove('plottest.out')
        except Exception:
            pass

    def make_savefile(self, plot_name):
        if PLOT_DIR:
            if not os.path.exists(PLOT_DIR):
                os.makedirs(PLOT_DIR)
            return os.path.join(PLOT_DIR, plot_name.lower().replace(' ', '_')), False
        else:
            return '', True

    def test_single_beachball(self):
        plot_name = "Single Beachball"
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, nodal_line=True, save_file=save_file, show=show)

    def test_multiple_mts_beachball_fail(self):
        with self.assertRaises(ValueError):
            MTplot(self.multiMTs, nodal_line=True)

    def test_nodal_line_plot(self):
        plot_name = 'Nodal line plot'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiMTs, 'faultplane', nodal_line=True, save_file=save_file, show=show)

    def test_lune_plot(self):
        plot_name = 'Lune Plot'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiMTs, 'lune', color='purple', save_file=save_file, show=show)

    def test_hudson_plot(self):
        plot_name = 'Hudson Plot'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiMTs, 'hudson', color='red', save_file=save_file, show=show)

    def test_riedesel_jordan_plot(self):
        plot_name = 'Riedesel Jordan Plot'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'riedeseljordan', color='purple', save_file=save_file, show=show)

    def test_riedesel_jordan_plot_multiple_mts_fail(self):
        with self.assertRaises(ValueError):
            MTplot(self.multiMTs, 'riedeseljordan', color='purple')

    def test_fault_plane_with_stations(self):
        plot_name = "Fault plane with stations"
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations,
               fault_plane=True, color='purple', save_file=save_file, show=show)

    def test_fault_plane_with_polarity_probability(self):
        plot_name = "Fault plane"
        self.stations['polarity'] = [0.8, 0.5, 0.3, 0.51, -0.95, 0.67, -0.95, 0.51, 0.78, 0.84, -0.65,
                                     0.83, 0.51, 0.68, 0.88, 0.91, 0.55, 0.51, 0.95, -0.66, 0.53, 0.66,
                                     -0.99, 0.99, 0.66, -0.58]

        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations,
               fault_plane=True, color='purple', save_file=save_file, show=show)

    def test_fault_plane_with_stations_station_dist(self):
        plot_name = 'Fault Plane and Stations with Station Dist'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations,
               station_distribution=self.station_distribution, fault_plane=True, color='purple', save_file=save_file, show=show)

    def test_stations_station_dist(self):
        plot_name = 'Stations and Station Dist Only'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations,
               station_distribution=self.station_distribution, fault_plane=False, color='purple', save_file=save_file, show=show)

    def test_fault_plane_with_station_dist(self):
        plot_name = 'Fault Plane no Stations with Station Dist'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations, show_stations=False,
               station_distribution=self.station_distribution, fault_plane=True, color='purple', save_file=save_file, show=show)

    def test_station_dist(self):
        plot_name = 'Station Dist Only'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations, show_stations=False,
               station_distribution=self.station_distribution, fault_plane=False, color='purple', save_file=save_file, show=show)

    def test_fault_plane_with_station_dist_2(self):
        plot_name = 'Fault Plane with Station Dist Only'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations={}, station_distribution=self.station_distribution, fault_plane=True,
               color='purple', save_file=save_file, show=show)

    def test_station_dist_2(self):
        plot_name = 'Station Dist Only 2'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations={}, station_distribution=self.station_distribution, fault_plane=False,
               color='purple', save_file=save_file, show=show)

    def test_fault_plane_with_prob_station_dist(self):
        self.station_distribution['probability'] = [305, 505]

        plot_name = 'Fault Plane with Prob Station Dist Only'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations, show_stations=False,
               station_distribution=self.station_distribution, fault_plane=True, color='purple', save_file=save_file, show=show)

    def test_prob_station_dist(self):
        self.station_distribution['probability'] = [305, 505]
        plot_name = 'Prob Station Dist Only'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'faultplane', stations=self.stations, show_stations=False,
               station_distribution=self.station_distribution, fault_plane=False, color='purple', save_file=save_file, show=show)
        print('With probabilities')
        print("Fault plane")

    def test_fault_plane_prob_with_prob_station_dist(self):
        self.station_distribution['probability'] = [305, 505]
        plot_name = 'Fault Plane Prob with Prob Station Dist'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiDCs, 'faultplane', probability=np.array([0.2, 0.6, 0.8, 0.1]), fault_plane=True, show_max_likelihood=True,
               show_mean=True, color='purple', save_file=save_file, show=show)

    def test_limited_fault_plane_prob_with_prob_station_dist(self):
        self.station_distribution['probability'] = [305, 505]
        plot_name = '2 Fault Planes with Prob Station Dist '
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiDCs, 'faultplane', probability=np.array([0.2, 0.6, 0.8, 0.1]), fault_plane=True, max_number_planes=2,
               show_max_likelihood=True, show_mean=True, color='purple', save_file=save_file, show=show)

    def test_lune_hist(self):
        plot_name = 'Lune Hist'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiMTs, 'lune', probability=np.array(
               [0.2, 0.6, 0.8, 0.1, 0.3, 0.7, 0.1, 0.8]), color='purple', save_file=save_file, show=show)

    def test_lune(self):
        plot_name = 'Lune'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'lune', color='purple', save_file=save_file, show=show)

    def test_hudson_hist(self):
        plot_name = 'Hudson'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiMTs, 'hudson', probability=np.array(
            [0.2, 0.6, 0.8, 0.1, 0.3, 0.7, 0.1, 0.8]), color='red', save_file=save_file, show=show)

    def test_hudson(self):
        plot_name = 'Hudson'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.singleMT, 'hudson', color='blue', save_file=save_file, show=show)

    def test_hudson_fault_plane_multi(self):
        plot_name = 'Hudson and Fault Plane'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot([self.multiMTs, self.multiDCs], ['hudson', 'fault_plane'], probability=[np.array([0.2, 0.6, 0.8, 0.1, 0.3, 0.7, 0.1, 0.8]),
                                                                                       np.array([0.2, 0.6, 0.8, 0.1])],
               show_max_likelihood=True, show_mean=True, color='red', save_file=save_file, show=show)

    def test_tape_hist(self):
        plot_name = 'Tape histogram'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiMTs, 'tape', hex_bin=0, save_file=save_file, show=show)

    def test_parameter_hist(self):
        plot_name = 'Parameter'
        print(plot_name)
        save_file, show = self.make_savefile(plot_name)
        MTplot(self.multiMTs, 'parameter', hex_bin=0, parameter='gamma', save_file=save_file, show=show)

    def test_output_file(self):
        MTplot(self.multiMTs, 'parameter', save_file='MTfit_plot_save_test.png',
               hex_bin=0, parameter='gamma', show=False)
        self.assertTrue(os.path.exists('MTfit_plot_save_test.png'), os.listdir('.'))
        try:
            os.remove('MTfit_plot_save_test.png')
        except Exception:
            pass



"""
Setup script for strike_slip
****************************
Call from command line as: sudo python setup.py install if on unix
or python setup.py install

"""
try:
    from setuptools import setup
except Exception:
    from distutils.core import setup


def setup_package():
    """
    setup_package()

    Function to call distutils setup to install the package to the default location for third party modules.

    Checks to see if required modules are installed and if not tries to install them.
    """
    kwargs = dict(name='model_sampling_strike_slip', version='1', author='David J Pugh', author_email='david.j.pugh@cantab.net',
                  packages=['model_sampling_strike_slip'],
                  package_dir={'model_sampling_strike_slip': '.'},
                  requires=['numpy', 'MTfit'],
                  install_requires=['numpy>=1.7.0'],
                  provides=['model_sampling_strike_slip'],
                  scripts=[],
                  description='model_sampling_strike_slip: Strike-slip random sampling for MTfit',
                  package_data={'strike_slip': []},)
    kwargs['entry_points'] = {'MTfit.sample_distribution': ['strike_slip= model_sampling_strike_slip:random_strike_slip']}
    setup(**kwargs)


if __name__ == "__main__":
    setup_package()



"""
**Restricted:  For Non-Commercial Use Only**
This code is protected intellectual property and is available solely for teaching
and non-commercially funded academic research purposes.

Applications for commercial use should be made to Schlumberger or the University of Cambridge.
"""
import sys
import multiprocessing
import json

from ._version import get_versions
__version__ = get_versions()['version']
from .run import MTfit  # noqa F401

__all__ = ['algorithms', 'convert', 'extensions', 'plot', 'utilities', 'inversion', 'probability', 'sampling']

del get_versions


def get_details():
    # Check if the different components are present
    c_extensions = []
    try:
        from MTfit.algorithms import cmarkov_chain_monte_carlo
        del cmarkov_chain_monte_carlo
        c_extensions.append('cmarkov_chain_monte_carlo')
    except Exception:
        pass
    try:
        from MTfit.probability import cprobability
        del cprobability
        c_extensions.append('cprobability')
    except Exception:
        pass
    try:
        from MTfit.convert import cmoment_tensor_conversion
        del cmoment_tensor_conversion
        c_extensions.append('cmoment_tensor_conversion')
    except Exception:
        pass
    try:
        from MTfit.extensions import cscatangle
        del cscatangle
        c_extensions.append('cscatangle')
    except Exception:
        pass
    dependency_versions = {}
    import numpy as np
    dependency_versions['numpy'] = np.__version__
    import scipy
    dependency_versions['scipy'] = scipy.__version__
    try:
        import matplotlib as mpl
        dependency_versions['matplotlib'] = mpl.__version__
    except ImportError:
        pass
    except AttributeError:
        dependency_versions['matplotlib'] = 'unknown'
    try:
        import cython
        dependency_versions['cython'] = cython.__version__
    except ImportError:
        pass
    except AttributeError:
        dependency_versions['cython'] = 'unknown'
    try:
        import pyqsub
        dependency_versions['pyqsub'] = pyqsub.__version__
    except ImportError:
        pass
    except AttributeError:
        dependency_versions['pyqsub'] = 'unknown'
    try:
        import sphinx
        dependency_versions['sphinx'] = sphinx.__version__
    except ImportError:
        pass
    except AttributeError:
        dependency_versions['sphinx'] = 'unknown'
    try:
        import h5py
        dependency_versions['h5py'] = h5py.__version__
    except ImportError:
        pass
    except AttributeError:
        dependency_versions['h5py'] = 'unknown'
    try:
        import hdf5storage
        dependency_versions['hdf5storage'] = hdf5storage.__version__
    except ImportError:
        pass
    except AttributeError:
        dependency_versions['hdf5storage'] = 'unknown'
    details = {'version': __version__,
               'c_extensions present': c_extensions,
               'platform': sys.platform,
               'num_threads': multiprocessing.cpu_count(),
               'python version': sys.version,
               'python version info': sys.version_info,
               'dependency info': dependency_versions}
    if sys.platform.startswith('win'):
        details['windows version'] = sys.getwindowsversion()
    if sys.version_info.major < 3:
        details['python version info'] = str(details['python version info'])
        if sys.platform.startswith('win'):
            details['windows version'] = str(details['windows version'])
    return details


def get_details_json():
    return json.dumps(get_details(), indent=4)



"""
test_cmoment_tensor_conversion.py
*****************

Tests for src/convert/cmoment_tensor_conversion.py
"""

import unittest

try:
    from MTfit.convert.cmoment_tensor_conversion import cMomentTensorConvertTestCase
except ImportError:
    class cMomentTensorConvertSkipTestCase(unittest.TestCase):

        def test_c_tape2mt6(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_E_tk(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_tk_uv(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_E_gd(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_N_sdr(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_TP_SDR(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_MT62TNPE(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_E2GD(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_TP2SDR(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")

        def test_c_SDR2SDR(self):
            raise unittest.SkipTest("No cmoment_tensor_conversion module")



"""
run.py
******
Core module for MTfit - handles all the command line parsing logic and calling the forward model based inversion.
"""

# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.


import os
import subprocess
import warnings
import traceback

# Test flag for running qsub flags in test
try:
    import pyqsub  # python module for cluster job submission using qsub.
    _PYQSUB = True
except Exception:
    _PYQSUB = False

from .inversion import Inversion
from .inversion import combine_mpi_output
from .utilities.extensions import get_extensions
from .extensions import default_pre_inversions
from .extensions import default_post_inversions
from .extensions import default_extensions
from .utilities.argparser import MTfit_parser


WARNINGS_MAP = {'a': 'always', 'd': 'default',
                'e': 'error', 'i': 'ignore',
                'm': 'module', 'o': 'once'}


ERROR_MESSAGE = """
**********************************************
Error running MTfit:

If this is a recurring error, please post an issue at https://github.com/djpugh/MTfit/issues/ (if one doesn't already exist)
including the traceback and the following information to help with diagnosis:

## Environment Information

```
{}
```

## Traceback

```
{}
```

**********************************************
"""


# Main MTfit function

def MTfit(data={}, data_file=False, location_pdf_file_path=False, algorithm='Time', parallel=True, n=0, phy_mem=8, dc=False, **kwargs):
    """
    Runs MTfit

    Creates an MTfit.inversion.Inversion object for the given arguments and runs the forward model based inversion.
    For a simple method of initialising the inversion use the command line approach (see MTfit docs).

    Args
        Data: Data dictionary (see MTfit.inversion.Inversion for structure).
        data_file: File or list of files which are pickled Data dictionaries.
        location_pdf_file_path: Path to angle scatter files or List of Angle Scatter Files (for Monte carlo marginalisation over location and model uncertainty)
            Can be generated from NonLinLoc *.scat files.
        algorithm:['Time'] Default search algorithm, for more information on the different algorithms see the MTfit.inversion.Inversion docstrings.
        parallel:['True'] Selects whether to run the inversion in parallel or on a single thread.
        n:[0] Number of threads to use, 0 defaults to all available threads reported by the system (from multiprocessing.cpu_count()).
        phy_mem:[8Gb] Estimated physical memory to use (used for determining array sizes, it is likely that more memory will be used, and if so no errors are forced).
        dc:[False] If true constrains the inversion to the double-couple space only.

    Keyword Arguments
        Test:[False] If true, runs unittests rather than the inversion.
        Other arguments to pass to the MTfit.inversion.Inversion object or for the algorithm (for more information see the MTfit.inversion.Inversion docstrings).

    Returns
        0

    """
    try:
        kwargs['data'] = data
        kwargs['data_file'] = data_file
        kwargs['location_pdf_file_path'] = location_pdf_file_path
        kwargs['algorithm'] = algorithm
        kwargs['parallel'] = parallel
        kwargs['n'] = n
        kwargs['phy_mem'] = phy_mem
        kwargs['dc'] = dc
        # GET PLUGINS
        pre_inversion_names, pre_inversions = get_extensions('MTfit.pre_inversion', default_pre_inversions)
        post_inversion_names, post_inversions = get_extensions('MTfit.post_inversion', default_post_inversions)
        extension_names, extensions = get_extensions('MTfit.extension', default_extensions)
        # Default extensions
        for ext in extensions.values():
            result = ext(**kwargs)
            if result != 1:
                return result
        # Check combine mpi output
        if kwargs.get('combine_mpi_output', False):
            combine_mpi_output(kwargs.get('path', ''), kwargs.get('output_format', 'matlab'), **kwargs)  # binary_file_version flag set here
            return 0
        if len(data) or (isinstance(data_file, str) and not os.path.isdir(data_file)) or isinstance(data_file, list):
            warnings.filterwarnings(WARNINGS_MAP[kwargs.get('warnings', 'd')])
            # Pre-inversion
            for plugin in pre_inversions.values():
                kwargs = plugin(**kwargs)
            if not kwargs.get('_mpi_call', False):
                print('Running MTfit.')
            # Effectively
            # inversion = Inversion(data, data_file, location_pdf_file_path, algorithm, parallel, n, phy_mem, dc, **kwargs)
            # but allowing the pre inversion plugin to change the kwargs
            inversion = Inversion(**kwargs)
            inversion.forward()
            if kwargs.get('dc_mt', False):
                # inversion = Inversion(data, data_file, location_pdf_file_path, algorithm, parallel, n, phy_mem, not dc, **kwargs)
                kwargs['dc'] = not kwargs['dc']
                inversion = Inversion(**kwargs)
                inversion.forward()
            # Post-inversion
            for plugin in post_inversions.values():
                plugin(**kwargs)
            return 0
    except Exception as e:
        from . import get_details_json
        print(ERROR_MESSAGE.format(get_details_json(), traceback.format_exc()))
        raise e


def run(args=None):
    """
    Runs inversion from command line arguments

    Runs the command line options parser and either submits job to cluster or runs the inversion on the local machine depending on the command line arguments

    Args
        args:[None] Input arguments - if not given, defaults to using sys.argv

    Returns
        0

    """
    options, options_map = MTfit_parser(args)
    if options['qsub'] and _PYQSUB:
        options_map['data_file'] = options_map['DATAFILE']
        options['singlethread'] = not options.pop('parallel')
        options['_mpi_call'] = options['mpi']
        if 'mcmc' in options['algorithm'].lower():
            options['max_samples'] = options['chain_length']
        return pyqsub.submit(options, options_map, __name__)
    else:
        for key in list(options.keys()):
            if 'qsub' in key:
                options.pop(key)
        if options['mpi'] and not options['_mpi_call']:
            try:
                # could add extra changeable MPI python handling here (and
                # elsewhere?)
                import mpi4py  # noqa F401
            except Exception:
                raise ImportError('MPI module mpi4py not found, unable to run in mpi')
            # restart python as mpirun
            options['_mpi_call'] = True
            print('Running MTfit using mpirun')
            optstring = pyqsub.make_optstr(options, options_map)
            mpiargs = ["mpirun", "-n", str(options['n']), "MTfit"]
            mpiargs.extend(optstring.split())
            return subprocess.call(mpiargs)
        return MTfit(**options)



"""unittest_utils
******************
Provides test functions for running and debugging unit tests.
"""

# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.

import unittest
import sys
import importlib
import traceback

import numpy as np


def get_extension_skip_if_args(module):
    reason = 'No C extension available'
    try:
        c_extension = importlib.import_module(module)
    except ImportError:
        c_extension = False
    except Exception:
        reason += '\n=======\nException loading C extension = \n{}\n=======\n'.format(traceback.format_exc())
        c_extension = False
    return (not c_extension, reason)


# class PythonOnly(object):

#     def __init__(self, *args):
#         self.c_extensions = []
#         self.original_values = []
#         for arg in args:
#             self.c_extensions.append(arg)
#             self.original_values = globals()[arg]

#     def __enter__(self, *args, **kwargs):
#         for extension in self.c_extensions:
#             globals()[extension] = False

#     def __exit__(self, *args, **kwargs):
#         for i, extension in enumerate(self.c_extensions):
#             globals()[extension] = self.original_values[i]


class TestCase(unittest.TestCase):

    def assertAlmostEqual(self, first, second, places=None, msg=None, delta=None):

        if isinstance(first, np.ndarray) or isinstance(second, np.ndarray):
            # Handle array and vectors
            if isinstance(second, (list, float, int)):
                second = np.array(second)
            if isinstance(first, (list, float, int)):
                first = np.array(first)
            if len([u for u in second.shape if u != 1]) == len([u for u in first.shape if u != 1]):
                if places is not None:
                    np.testing.assert_array_almost_equal(np.array(first).squeeze(), np.array(second).squeeze(), places)
                else:
                    np.testing.assert_array_almost_equal(np.array(first).squeeze(), np.array(second).squeeze())
            else:
                if places is not None:
                    np.testing.assert_array_almost_equal(first, second, places)
                else:
                    np.testing.assert_array_almost_equal(first, second)
            return
        elif isinstance(first, dict) and isinstance(second, dict):
            self.assertEqual(sorted(first.keys()), sorted(second.keys()), 'Dictionary keys do not match')
            for key in first.keys():
                self.assertAlmostEqual(first[key], second[key])
            return
        elif isinstance(first, list) and isinstance(second, list):
            super(TestCase, self).assertAlmostEqual(set(first), set(second), msg, delta)
        else:
            super(TestCase, self).assertAlmostEqual(first, second, places, msg, delta)

    def assertAlmostEquals(self, *args, **kwargs):
        self.assertAlmostEqual(*args, **kwargs)

    def assertEqual(self, first, second, msg=None):
        if isinstance(first, np.ndarray) or isinstance(second, np.ndarray):
            # Handle matrix and vector

            np.testing.assert_array_equal(first, second)
            return
        elif isinstance(first, dict) and isinstance(second, dict):
            self.assertEqual(
                sorted(first.keys()), sorted(second.keys()), 'Dictionary keys do not match')
            for key in first.keys():
                self.assertEqual(first[key], second[key])
            return
        elif isinstance(first, list) and isinstance(second, list):
            super(TestCase, self).assertEqual(first, second, msg)

        else:
            super(TestCase, self).assertAlmostEqual(first, second, msg)

    def assertEquals(self, *args, **kwargs):
        self.assertEqual(*args, **kwargs)

    def assertVectorEquals(self, first, second, *args):
        try:
            first_norm = np.sqrt(np.sum(np.multiply(first, first)))
            second_norm = np.sqrt(np.sum(np.multiply(second, second)))
            return self.assertAlmostEquals(first/first_norm, second/second_norm, *args)
        except AssertionError as e1:
            try:
                return self.assertAlmostEquals(-first/first_norm, second/second_norm, *args)
            except AssertionError as e2:
                if sys.version_info.major <= 2 and sys.version_info.minor <= 6:
                    raise AssertionError(e1.message+' or '+e2.message)
                else:
                    raise AssertionError('{} or {}'.format(e1.args, e2.args))



import unittest
import sys
import glob
import os
import tempfile
import shutil

from MTfit.extensions.scatangle import parse_scatangle
from MTfit.extensions import scatangle
from MTfit.utilities import C_EXTENSION_FALLBACK_LOG_MSG
from MTfit.utilities.unittest_utils import get_extension_skip_if_args

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock


C_EXTENSIONS = get_extension_skip_if_args('MTfit.extension.cscatangle')


class PythonOnly(object):

    def __enter__(self, *args, **kwargs):
        self.cscatangle = scatangle.cscatangle
        scatangle.cscatangle = False

    def __exit__(self, *args, **kwargs):
        scatangle.cscatangle = self.cscatangle


class ScatangleTestCase(unittest.TestCase):

    def setUp(self):
        self.cwd = os.getcwd()
        if sys.version_info >= (3, 0):
            self.tempdir = tempfile.TemporaryDirectory()
            os.chdir(self.tempdir.name)
        else:
            self.tempdir = tempfile.mkdtemp()
            os.chdir(self.tempdir)
        self.existing_scatangle_files = glob.glob('*.scatangle')

    def tearDown(self):
        for fname in glob.glob('*.scatangle'):
            if fname not in self.existing_scatangle_files:
                try:
                    os.remove(fname)
                except Exception:
                    print('Cannot remove {}'.format(fname))
        import gc
        try:
            os.remove('test.scatangle')
        except Exception:
            pass
        os.chdir(self.cwd)
        if sys.version_info >= (3, 0):
            self.tempdir.cleanup()
        else:
            try:
                shutil.rmtree(self.tempdir)
            except:
                pass
        gc.collect()

    def station_angles(self):
        out = "504.7\n"
        out += "S0271   231.1   154.7\n"
        out += "S0649   42.9    109.7\n"
        out += "S0484   21.2    145.4\n"
        out += "S0263   256.4   122.7\n"
        out += "S0142   197.4   137.6\n"
        out += "S0244   229.7   148.1\n"
        out += "S0415   75.6    122.8\n"
        out += "S0065   187.5   126.1\n"
        out += "S0362   85.3    128.2\n"
        out += "S0450   307.5   137.7\n"
        out += "S0534   355.8   138.2\n"
        out += "S0641   14.7    120.2\n"
        out += "S0155   123.5   117\n"
        out += "S0162   231.8   127.5\n"
        out += "S0650   45.9    108.2\n"
        out += "S0195   193.8   147.3\n"
        out += "S0517   53.7    124.2\n"
        out += "S0004   218.4   109.8\n"
        out += "S0588   12.9    128.6\n"
        out += "S0377   325.5   165.3\n"
        out += "S0618   29.4    120.5\n"
        out += "S0347   278.9   149.5\n"
        out += "S0529   326.1   131.7\n"
        out += "S0083   223.7   118.2\n"
        out += "S0595   42.6    117.8\n"
        out += "S0236   253.6   118.6\n"
        out += '\n'
        out += "504.7\n"
        out += "S0271   230.9   154.8\n"
        out += "S0649   43      109.8\n"
        out += "S0484   21.3    145.4\n"
        out += "S0263   256.4   122.8\n"
        out += "S0142   197.3   137.6\n"
        out += "S0244   229.6   148.1\n"
        out += "S0415   75.7    122.8\n"
        out += "S0065   187.4   126.1\n"
        out += "S0362   85.3    128.2\n"
        out += "S0450   307.5   137.8\n"
        out += "S0534   355.7   138.3\n"
        out += "S0641   14.8    120.3\n"
        out += "S0155   123.5   117.1\n"
        out += "S0162   231.7   127.6\n"
        out += "S0650   45.9    108.3\n"
        out += "S0195   193.6   147.3\n"
        out += "S0517   53.7    124.2\n"
        out += "S0004   218.3   109.9\n"
        out += "S0588   13      128.7\n"
        out += "S0377   325.7   165.4\n"
        out += "S0618   29.5    120.5\n"
        out += "S0347   278.9   149.6\n"
        out += "S0529   326.1   131.8\n"
        out += "S0083   223.7   118.2\n"
        out += "S0595   42.7    117.9\n"
        out += "S0236   253.5   118.7\n"
        out += "\n"
        return out

    @mock.patch('MTfit.extensions.scatangle.logger')
    def test_parser_scatangle(self, logger, parser=parse_scatangle, *args):
        with PythonOnly():
            open('test.scatangle', 'w').write(self.station_angles())
            A, B = parser('test.scatangle')
            self.assertEqual(B, [504.7, 504.7])
            self.assertEqual(len(A), 2)
            self.assertEqual(sorted(A[0].keys()), ['Azimuth', 'Name', 'TakeOffAngle'])
            A, B = parser('test.scatangle', bin_size=1)
            self.assertEqual(B, [1009.4])
            self.assertEqual(len(A), 1)
            self.assertEqual(sorted(A[0].keys()), ['Azimuth', 'Name', 'TakeOffAngle'])
            self.assertIn(mock.call('Python code used'), logger.info.call_args_list)
            self.assertNotIn(mock.call('C code used'), logger.info.call_args_list)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.extensions.scatangle.logger')
    def test_parser_scatangle_cython(self, logger, parser=parse_scatangle, *args):
        open('test.scatangle', 'w').write('\n'.join([self.station_angles() for i in range(40)]))
        AC, BC = parser('test.scatangle', bin_size=1)
        self.assertNotIn(mock.call('Python code used'), logger.info.call_args_list)
        self.assertIn(mock.call('C code used'), logger.info.call_args_list)
        APy, BPy = parser('test.scatangle', bin_size=1, _use_c=False)
        self.assertEqual(len(APy), 1)
        self.assertEqual(len(AC), 1)
        self.assertEqual(BPy, BC)
        os.remove('test.scatangle')
